# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.

Interpolation of fitted surrogate model.

JJ 2016-11-30
"""

from __future__ import division
from __future__ import absolute_import

import numpy as np  # needed by interpolate_fit() in Python API

cimport wlsqm.fitter.defs     as defs      # C constants
cimport wlsqm.fitter.polyeval as polyeval  # evaluation of Taylor expansions and general polynomials

####################################################
# Python API
####################################################

def interpolate_fit( xi, double[::1] fi, int dimension, int order, x, int diff=0 ):
    """def interpolate_fit( xi, double[::1] fi, int dimension, int order, x, int diff=0 ):

Interpolate the fit to given points x.

This calls the C API to interpolate the model; thus the result is guaranteed to be identical
to the internal model interpolation performed during fitting with iterative refinement.

Fused multiply-add (fma()) is used to reduce roundoff.

xi    : in, the origin of the fit; (x0,y0,z0) (double[:]) in 3D, (x0,y0) (double[:]) in 2D, or (x0) (double) in 1D
fi    : in, fit coefficients (double[::1])
order : in, degree of the surrogate polynomial (int)
x     : in, points to which to interpolate the model; x[k,:] = (x,y,z) in 3D, x[k,:] = (x,y) in 2D, or x[:] = (x) in 1D
diff  : in, one of the i1_* (if dimension==1), i2_* (if dimension==2) or i3_* (if dimension==3) constants in wlsqm.fitter.defs.

        If dimension==1:

        i1_F   : interpolate function value f (default)
        i1_X   : interpolate df/dx
        i1_X2  : interpolate d2f/dx2
        i1_X3  : ...
        i1_X4

        If dimension==2:

        i2_F   : interpolate function value f (default)
        i2_X   : interpolate df/dx
        i2_Y   : interpolate df/dy
        i2_X2  : interpolate d2f/dx2
        i2_XY  : interpolate d2f/dxdy
        i2_Y2  : interpolate d2f/dy2
        i2_X3  : ...
        i2_X2Y
        i2_XY2
        i2_Y3
        i2_X4
        i2_X3Y
        i2_X2Y2
        i2_XY3
        i2_Y4

        For dimension==3, see wlsqm.fitter.defs (there are 35 constants for the 3D case).

        Derivatives of order up to that of the model order are available.
        E.g. if order=2, then up to second-order derivatives are available.

        This can be used to obtain the derivatives at a general point
        (i.e. not only at the origin of the fit; that information
         is already contained in the coefficients, since the model
         is a polynomial).

        Note that in general, the higher the derivative order,
        the less accurate the result will be. This is due to the
        fact that the fitting error mostly concentrates into
        the higher-order terms.

Return value : rank-1 array, function value at each x.
"""
    if dimension not in [1,2,3]:
        raise ValueError( "dimension must be 1, 2 or 3; got %d" % dimension )
    if order not in [0,1,2,3,4]:
        raise ValueError( "order must be 0, 1, 2, 3 or 4; got %d" % order )

    cdef double[::1] xiManyD
    cdef double[::view.generic,::view.contiguous] xManyD

    cdef double xi1D
    cdef double[::view.generic] x1D

    cdef int nx
    x  = np.atleast_1d( x )
    nx = x.shape[0]

    cdef double[::1] out = np.empty( (nx,), dtype=np.float64 )

    cdef double nan = 0./0.
    cdef double x0, y0, z0
    if dimension >= 2:
        x0     = xi[0]
        y0     = xi[1]
        z0     = xi[2]  if dimension == 3  else  nan
        xManyD = x
        x1D    = None

    else: # dimension == 1:
        x0     = xi
        y0     = nan
        z0     = nan
        xManyD = None
        x1D    = x

    # TODO: to save a few bytes, could use "no2" (currently not saved), not "no". If order >= 1, no2 = number_of_dofs(dimension, order-1); if order == 0, the workspace fi2 is not needed.
    cdef int no = infra.number_of_dofs(dimension, order)
    cdef double[::1] fi2 = np.empty( (no,), dtype=np.float64 )

    # We don't have a Case instance here, since we are coming from the Python level.
    # Now we must create one, or at least populate the required fields manually.
    #
    cdef infra.Case case
    with nogil:
        case.dimension = dimension
        case.order     = order
        case.no        = no
        case.xi        = x0
        case.yi        = y0
        case.zi        = z0
        case.fi        = &fi[0]
        case.fi2       = &fi2[0]  # if interpolating a derivative, this work space is needed (if we had a Case object, the work space would have been created by Case_allocate())
        interpolate_nD( &case, xManyD, x1D, &out[0], diff )

    return out


def lambdify_fit( xi, fi, dimension, order, diff=0 ):  # Python only
    """def lambdify_fit( xi, fi, dimension, order, diff=0 ):

Create a Python lambda that interpolates a fitted surrogate model.

If dimension = 3, the output is lambda x,y,z : ... that interpolates the model to (x,y,z).
If dimension = 2, the output is lambda x,y   : ... that interpolates the model to (x,y).
If dimension = 1, the output is lambda x     : ... that interpolates the model to x.

The returned lambda is vectorized in the sense that it can be called with an np.array of coordinates.
In the 2D and 3D cases, the lengths of the arguments must be compatible (i.e. either all of the same length or scalars).

Note that the same xi, dimension and order must be passed here as for the model fitting.

fi[] must be the coefficients output by one of the fitting functions (see wlsqm.fitter.simple.fit_*()).

The parameter "diff" allows lambdifying a derivative of the model instead of the function itself.
See interpolate_fit() for the valid values. The default is to lambdify the function itself.
"""
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    if dimension not in [1,2,3]:
        raise ValueError( "dimension must be 1, 2 or 3; got %d" % dimension )
    if order not in [0,1,2,3,4]:
        raise ValueError( "order must be 0, 1, 2, 3 or 4; got %d" % order )

    # Make an adapter that wraps interpolate_fit(), freezing in the given xi, fi, dimension, order.
    #
    if dimension == 3:
        def model(x, y, z):
            if np.shape(y) != np.shape(x)  or  np.shape(z) != np.shape(x):
                raise ValueError("In model() (generated by wlsqm.fitter.interp.lambdify_fit()): x, y and z must be of the same shape; got shape(x) = %s, shape(y) = %s, shape(z) = %s" % (np.shape(x), np.shape(y), np.shape(z)))

            # force x and y into arrays of the same shape
            #
            nx = np.size(x)
            ny = np.size(y)
            if nx > 1  and  ny == 1:
                y = np.ones_like(x) * y
            elif nx == 1  and  ny > 1:
                x = np.ones_like(y) * x
            elif nx == 1  and  ny == 1:
                x = np.atleast_1d(x)
                y = np.atleast_1d(y)
            # now x and y are of the same shape, and both arrays; process z
            nz = np.size(z)
            if nz == 1:
                z = np.ones_like(x) * z

            # make a single array of shape (nx,2)
            #
            # xy[j,:] contains (x,y) for the jth point
            #
            xy  = np.empty( (np.size(x),3), dtype=np.float64 )
            shp = np.shape(x)
            xy[:,0] = np.reshape(x, -1)
            xy[:,1] = np.reshape(y, -1)
            xy[:,2] = np.reshape(z, -1)

            return np.reshape( interpolate_fit( xi, fi, 3, order, xy, diff ), shp )

    elif dimension == 2:
        def model(x, y):
            if np.shape(y) != np.shape(x):
                raise ValueError("In model() (generated by wlsqm.fitter.interp.lambdify_fit()): x and y must be of the same shape; got shape(x) = %s, shape(y) = %s" % (np.shape(x), np.shape(y)))

            # force x and y into arrays of the same shape
            #
            nx = np.size(x)
            ny = np.size(y)
            if nx > 1  and  ny == 1:
                y = np.ones_like(x) * y
            elif nx == 1  and  ny > 1:
                x = np.ones_like(y) * x
            elif nx == 1  and  ny == 1:
                x = np.atleast_1d(x)
                y = np.atleast_1d(y)

            # make a single array of shape (nx,2)
            #
            # xy[j,:] contains (x,y) for the jth point
            #
            xy  = np.empty( (np.size(x),2), dtype=np.float64 )
            shp = np.shape(x)
            xy[:,0] = np.reshape(x, -1)
            xy[:,1] = np.reshape(y, -1)

            return np.reshape( interpolate_fit( xi, fi, 2, order, xy, diff ), shp )

    else: # dimension == 1:
        def model(x):
            tmp = np.atleast_1d(x)
            return np.asanyarray( interpolate_fit( xi, fi, 1, order, tmp, diff ) )  # the asanyarray() is needed to convert memoryviewslice --> np.array

    return model


####################################################
# C API
####################################################

# Interpolation of the model and its derivatives.

# Adapter that takes in parameters for 1D, 2D and 3D cases, calls the appropriate routine and passes on the relevant parameters.
#
# This is used to work around the static typing in C to allow a single dimension-agnostic implementation for functions needing model interpolation.
#
cdef int interpolate_nD( infra.Case* case, double[::view.generic,::view.contiguous] xManyD, double[::view.generic] x1D, double* out, int diff ) nogil:
    if case.dimension == 3:
        return interpolate_3D( case, xManyD, out, diff )
    elif case.dimension == 2:
        return interpolate_2D( case, xManyD, out, diff )
    else: # case.dimension == 1:
        return interpolate_1D( case, x1D, out, diff )


# case  : infra.Case object with the necessary metadata
#     xi    : in, x0 of the origin (x0,y0,z0) of the fit
#     yi    : in, y0 of the origin (x0,y0,z0) of the fit
#     zi    : in, z0 of the origin (x0,y0,z0) of the fit
#     order : in, degree of the surrogate polynomial
#     fi    : in, fit coefficients
# x     : in, x[k,:] = (x,y,z) to which to interpolate the model
# out   : out, function values. Must be allocated by caller. Must have as nk elements (one element for each point x).
# diff  : in, one of the i3_* constants in wlsqm.fitter.defs.
#
# Note that the same xi and order must be passed here as for the model fitting.
# fi[] must contain the output the model fitting.
#
cdef int interpolate_3D( infra.Case* case, double[::view.generic,::view.contiguous] x, double* out, int diff ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    cdef int order = case.order
    cdef int k,n

    # Interpolate function value? (the most common case)
    if diff == defs.i3_F_c:
        return polyeval.taylor_3D( order, case.fi, case.xi, case.yi, case.zi, x, out )

    # Interpolate a derivative of an order higher than the order of the model?
    #
    # The result is identically zero, so we special-case it here.
    #
    # Note that:
    #   - The constants used as values of "diff" are ordered in increasing order of derivatives
    #   - infra.number_of_dofs() returns one-past-end indices for each order of derivatives
    #   - This information is cached into case.no
    #
    elif diff >= case.no:  # infra.number_of_dofs( case.dimension, case.order ):
        n = x.shape[0]
        for k in range(n):
            out[k] = 0.
        return 0  # success

    # Else this derivative may be nonzero, handle it normally.

    # We fill in the coefficient array of the derivative polynomial,
    # shifting the coefficients to their new positions in the array
    # and accounting for exponents that "drop down" in the differentiation.
    #
    # We must also account for the constant factors in the Taylor expansion,
    # since fi[] contains only the DOF value data (and the DOFs are the
    # function value and derivatives at the point xi).
    #
    # Then we use the generic polynomial evaluator.

    cdef double* fi  = case.fi
    cdef double* fi2 = case.fi2  # work space

    # --------------- 1st order derivatives ---------------
    # if we get here, model order >= 1

    if diff == defs.i3_X_c:
        fi2[ defs.i3_F_c ]       = 1.*fi[ defs.i3_X_c    ]
        if order >= 2:
            fi2[ defs.i3_X_c   ] = 2.*fi[ defs.i3_X2_c   ] * 0.5
            fi2[ defs.i3_Y_c   ] = 1.*fi[ defs.i3_XY_c   ]
            fi2[ defs.i3_Z_c   ] = 1.*fi[ defs.i3_XZ_c   ]
        if order >= 3:
            fi2[ defs.i3_X2_c  ] = 3.*fi[ defs.i3_X3_c   ] * onesixth
            fi2[ defs.i3_XY_c  ] = 2.*fi[ defs.i3_X2Y_c  ] * 0.5
            fi2[ defs.i3_Y2_c  ] = 1.*fi[ defs.i3_XY2_c  ] * 0.5
            fi2[ defs.i3_YZ_c  ] = 1.*fi[ defs.i3_XYZ_c  ]
            fi2[ defs.i3_Z2_c  ] = 1.*fi[ defs.i3_XZ2_c  ] * 0.5
            fi2[ defs.i3_XZ_c  ] = 2.*fi[ defs.i3_X2Z_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X3_c  ] = 4.*fi[ defs.i3_X4_c   ] * one24th
            fi2[ defs.i3_X2Y_c ] = 3.*fi[ defs.i3_X3Y_c  ] * onesixth
            fi2[ defs.i3_XY2_c ] = 2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_Y3_c  ] = 1.*fi[ defs.i3_XY3_c  ] * onesixth
            fi2[ defs.i3_Y2Z_c ] = 1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_YZ2_c ] = 1.*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_Z3_c  ] = 1.*fi[ defs.i3_XZ3_c  ] * onesixth
            fi2[ defs.i3_XZ2_c ] = 2.*fi[ defs.i3_X2Z2_c ] * 0.25
            fi2[ defs.i3_X2Z_c ] = 3.*fi[ defs.i3_X3Z_c  ] * onesixth
            fi2[ defs.i3_XYZ_c ] = 2.*fi[ defs.i3_X2YZ_c ] * 0.5

        return polyeval.general_3D( order-1, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y_c:
        fi2[ defs.i3_F_c ]       = 1.*fi[ defs.i3_Y_c    ]
        if order >= 2:
            fi2[ defs.i3_X_c   ] = 1.*fi[ defs.i3_XY_c   ]
            fi2[ defs.i3_Y_c   ] = 2.*fi[ defs.i3_Y2_c   ] * 0.5
            fi2[ defs.i3_Z_c   ] = 1.*fi[ defs.i3_YZ_c   ]
        if order >= 3:
            fi2[ defs.i3_X2_c  ] = 1.*fi[ defs.i3_X2Y_c  ] * 0.5
            fi2[ defs.i3_XY_c  ] = 2.*fi[ defs.i3_XY2_c  ] * 0.5
            fi2[ defs.i3_Y2_c  ] = 3.*fi[ defs.i3_Y3_c   ] * onesixth
            fi2[ defs.i3_YZ_c  ] = 2.*fi[ defs.i3_Y2Z_c  ] * 0.5
            fi2[ defs.i3_Z2_c  ] = 1.*fi[ defs.i3_YZ2_c  ] * 0.5
            fi2[ defs.i3_XZ_c  ] = 1.*fi[ defs.i3_XYZ_c  ]
        if order >= 4:
            fi2[ defs.i3_X3_c  ] = 1.*fi[ defs.i3_X3Y_c  ] * onesixth
            fi2[ defs.i3_X2Y_c ] = 2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_XY2_c ] = 3.*fi[ defs.i3_XY3_c  ] * onesixth
            fi2[ defs.i3_Y3_c  ] = 4.*fi[ defs.i3_Y4_c   ] * one24th
            fi2[ defs.i3_Y2Z_c ] = 3.*fi[ defs.i3_Y3Z_c  ] * onesixth
            fi2[ defs.i3_YZ2_c ] = 2.*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_Z3_c  ] = 1.*fi[ defs.i3_YZ3_c  ] * onesixth
            fi2[ defs.i3_XZ2_c ] = 1.*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_X2Z_c ] = 1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_XYZ_c ] = 2.*fi[ defs.i3_XY2Z_c ] * 0.5

        return polyeval.general_3D( order-1, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Z_c:
        fi2[ defs.i3_F_c ]       = 1.*fi[ defs.i3_Z_c    ]
        if order >= 2:
            fi2[ defs.i3_X_c   ] = 1.*fi[ defs.i3_XZ_c   ]
            fi2[ defs.i3_Y_c   ] = 1.*fi[ defs.i3_YZ_c   ]
            fi2[ defs.i3_Z_c   ] = 2.*fi[ defs.i3_Z2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i3_X2_c  ] = 1.*fi[ defs.i3_X2Z_c  ] * 0.5
            fi2[ defs.i3_XY_c  ] = 1.*fi[ defs.i3_XYZ_c  ]
            fi2[ defs.i3_Y2_c  ] = 1.*fi[ defs.i3_Y2Z_c  ] * 0.5
            fi2[ defs.i3_YZ_c  ] = 2.*fi[ defs.i3_YZ2_c  ] * 0.5
            fi2[ defs.i3_Z2_c  ] = 3.*fi[ defs.i3_Z3_c   ] * onesixth
            fi2[ defs.i3_XZ_c  ] = 2.*fi[ defs.i3_XZ2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X3_c  ] = 1.*fi[ defs.i3_X3Z_c  ] * onesixth
            fi2[ defs.i3_X2Y_c ] = 1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_XY2_c ] = 1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_Y3_c  ] = 1.*fi[ defs.i3_Y3Z_c  ] * onesixth
            fi2[ defs.i3_Y2Z_c ] = 2.*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_YZ2_c ] = 3.*fi[ defs.i3_YZ3_c  ] * onesixth
            fi2[ defs.i3_Z3_c  ] = 4.*fi[ defs.i3_Z4_c   ] * one24th
            fi2[ defs.i3_XZ2_c ] = 3.*fi[ defs.i3_XZ3_c  ] * onesixth
            fi2[ defs.i3_X2Z_c ] = 2.*fi[ defs.i3_X2Z2_c ] * 0.25
            fi2[ defs.i3_XYZ_c ] = 2.*fi[ defs.i3_XYZ2_c ] * 0.5

        return polyeval.general_3D( order-1, fi2, case.xi, case.yi, case.zi, x, out )

    # --------------- 2nd order derivatives ---------------
    # if we get here, model order >= 2

    elif diff == defs.i3_X2_c:
        fi2[ defs.i3_F_c ]      = 1.*2.*fi[ defs.i3_X2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 2.*3.*fi[ defs.i3_X3_c   ] * onesixth
            fi2[ defs.i3_Y_c  ] = 1.*2.*fi[ defs.i3_X2Y_c  ] * 0.5
            fi2[ defs.i3_Z_c  ] = 1.*2.*fi[ defs.i3_X2Z_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 3.*4.*fi[ defs.i3_X4_c   ] * one24th
            fi2[ defs.i3_XY_c ] = 2.*3.*fi[ defs.i3_X3Y_c  ] * onesixth
            fi2[ defs.i3_Y2_c ] = 1.*2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_YZ_c ] = 1.*2.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_Z2_c ] = 1.*2.*fi[ defs.i3_X2Z2_c ] * 0.25
            fi2[ defs.i3_XZ_c ] = 2.*3.*fi[ defs.i3_X3Z_c  ] * onesixth

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XY_c:
        fi2[ defs.i3_F_c ]      = 1.*1.*fi[ defs.i3_XY_c   ]
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 2.*1.*fi[ defs.i3_X2Y_c  ] * 0.5
            fi2[ defs.i3_Y_c  ] = 1.*2.*fi[ defs.i3_XY2_c  ] * 0.5
            fi2[ defs.i3_Z_c  ] = 1.*1.*fi[ defs.i3_XYZ_c  ]
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 3.*1.*fi[ defs.i3_X3Y_c  ] * onesixth
            fi2[ defs.i3_XY_c ] = 2.*2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_Y2_c ] = 1.*3.*fi[ defs.i3_XY3_c  ] * onesixth
            fi2[ defs.i3_YZ_c ] = 1.*2.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_Z2_c ] = 1.*1.*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_XZ_c ] = 2.*1.*fi[ defs.i3_X2YZ_c ] * 0.5

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y2_c:
        fi2[ defs.i3_F_c ]      = 1.*2.*fi[ defs.i3_Y2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 1.*2.*fi[ defs.i3_XY2_c  ] * 0.5
            fi2[ defs.i3_Y_c  ] = 2.*3.*fi[ defs.i3_Y3_c   ] * onesixth
            fi2[ defs.i3_Z_c  ] = 1.*2.*fi[ defs.i3_Y2Z_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 1.*2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_XY_c ] = 2.*3.*fi[ defs.i3_XY3_c  ] * onesixth
            fi2[ defs.i3_Y2_c ] = 3.*4.*fi[ defs.i3_Y4_c   ] * one24th
            fi2[ defs.i3_YZ_c ] = 2.*3.*fi[ defs.i3_Y3Z_c  ] * onesixth
            fi2[ defs.i3_Z2_c ] = 1.*2.*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_XZ_c ] = 1.*2.*fi[ defs.i3_XY2Z_c ] * 0.5

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_YZ_c:
        fi2[ defs.i3_F_c ]      = 1.*1.*fi[ defs.i3_YZ_c   ]
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 1.*1.*fi[ defs.i3_XYZ_c  ]
            fi2[ defs.i3_Y_c  ] = 2.*1.*fi[ defs.i3_Y2Z_c  ] * 0.5
            fi2[ defs.i3_Z_c  ] = 1.*2.*fi[ defs.i3_YZ2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 1.*1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_XY_c ] = 2.*1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_Y2_c ] = 3.*1.*fi[ defs.i3_Y3Z_c  ] * onesixth
            fi2[ defs.i3_YZ_c ] = 2.*2.*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_Z2_c ] = 1.*3.*fi[ defs.i3_YZ3_c  ] * onesixth
            fi2[ defs.i3_XZ_c ] = 1.*2.*fi[ defs.i3_XYZ2_c ] * 0.5

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Z2_c:
        fi2[ defs.i3_F_c ]      = 1.*2.*fi[ defs.i3_Z2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 1.*2.*fi[ defs.i3_XZ2_c  ] * 0.5
            fi2[ defs.i3_Y_c  ] = 1.*2.*fi[ defs.i3_YZ2_c  ] * 0.5
            fi2[ defs.i3_Z_c  ] = 2.*3.*fi[ defs.i3_Z3_c   ] * onesixth
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 1.*2.*fi[ defs.i3_X2Z2_c ] * 0.25
            fi2[ defs.i3_XY_c ] = 1.*2.*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_Y2_c ] = 1.*2.*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_YZ_c ] = 2.*3.*fi[ defs.i3_YZ3_c  ] * onesixth
            fi2[ defs.i3_Z2_c ] = 3.*4.*fi[ defs.i3_Z4_c   ] * one24th
            fi2[ defs.i3_XZ_c ] = 2.*3.*fi[ defs.i3_XZ3_c  ] * onesixth

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XZ_c:
        fi2[ defs.i3_F_c ]      = 1.*1.*fi[ defs.i3_XZ_c   ]
        if order >= 3:
            fi2[ defs.i3_X_c  ] = 2.*1.*fi[ defs.i3_X2Z_c  ] * 0.5
            fi2[ defs.i3_Y_c  ] = 1.*1.*fi[ defs.i3_XYZ_c  ]
            fi2[ defs.i3_Z_c  ] = 1.*2.*fi[ defs.i3_XZ2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X2_c ] = 3.*1.*fi[ defs.i3_X3Z_c  ] * onesixth
            fi2[ defs.i3_XY_c ] = 2.*1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_Y2_c ] = 1.*1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_YZ_c ] = 1.*2.*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_Z2_c ] = 1.*3.*fi[ defs.i3_XZ3_c  ] * onesixth
            fi2[ defs.i3_XZ_c ] = 2.*2.*fi[ defs.i3_X2Z2_c ] * 0.25

        return polyeval.general_3D( order-2, fi2, case.xi, case.yi, case.zi, x, out )

    # --------------- 3rd order derivatives ---------------
    # if we get here, model order >= 3

    elif diff == defs.i3_X3_c:
        fi2[ defs.i3_F_c ]     = 1.*2.*3.*fi[ defs.i3_X3_c  ] * onesixth
        if order >= 4:
            fi2[ defs.i3_X_c ] = 2.*3.*4.*fi[ defs.i3_X4_c  ] * one24th
            fi2[ defs.i3_Y_c ] = 1.*2.*3.*fi[ defs.i3_X3Y_c ] * onesixth
            fi2[ defs.i3_Z_c ] = 1.*2.*3.*fi[ defs.i3_X3Z_c ] * onesixth

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X2Y_c:
        fi2[ defs.i3_F_c ]     = (1.*2.)*1.*fi[ defs.i3_X2Y_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = (2.*3.)*1.*fi[ defs.i3_X3Y_c  ] * onesixth
            fi2[ defs.i3_Y_c ] = (1.*2.)*2.*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_Z_c ] = (1.*2.)*1.*fi[ defs.i3_X2YZ_c ] * 0.5

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XY2_c:
        fi2[ defs.i3_F_c ]     = 1.*(1.*2.)*fi[ defs.i3_XY2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = 2.*(1.*2.)*fi[ defs.i3_X2Y2_c ] * 0.25
            fi2[ defs.i3_Y_c ] = 1.*(2.*3.)*fi[ defs.i3_XY3_c  ] * onesixth
            fi2[ defs.i3_Z_c ] = 1.*(1.*2.)*fi[ defs.i3_XY2Z_c ] * 0.5

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y3_c:
        fi2[ defs.i3_F_c ]     = 1.*2.*3.*fi[ defs.i3_Y3_c  ] * onesixth
        if order >= 4:
            fi2[ defs.i3_X_c ] = 1.*2.*3.*fi[ defs.i3_XY3_c ] * onesixth
            fi2[ defs.i3_Y_c ] = 2.*3.*4.*fi[ defs.i3_Y4_c  ] * one24th
            fi2[ defs.i3_Z_c ] = 1.*2.*3.*fi[ defs.i3_Y3Z_c ] * onesixth

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y2Z_c:
        fi2[ defs.i3_F_c ]     = (1.*2.)*1.*fi[ defs.i3_Y2Z_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = (1.*2.)*1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_Y_c ] = (2.*3.)*1.*fi[ defs.i3_Y3Z_c  ] * onesixth
            fi2[ defs.i3_Z_c ] = (1.*2.)*2.*fi[ defs.i3_Y2Z2_c ] * 0.25

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_YZ2_c:
        fi2[ defs.i3_F_c ]     = 1.*(1.*2.)*fi[ defs.i3_YZ2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = 1.*(1.*2.)*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_Y_c ] = 2.*(1.*2.)*fi[ defs.i3_Y2Z2_c ] * 0.25
            fi2[ defs.i3_Z_c ] = 1.*(2.*3.)*fi[ defs.i3_YZ3_c  ] * onesixth

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Z3_c:
        fi2[ defs.i3_F_c ]     = 1.*2.*3.*fi[ defs.i3_Z3_c  ] * onesixth
        if order >= 4:
            fi2[ defs.i3_X_c ] = 1.*2.*3.*fi[ defs.i3_XZ3_c ] * onesixth
            fi2[ defs.i3_Y_c ] = 1.*2.*3.*fi[ defs.i3_YZ3_c ] * onesixth
            fi2[ defs.i3_Z_c ] = 2.*3.*4.*fi[ defs.i3_Z4_c  ] * one24th

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XZ2_c:
        fi2[ defs.i3_F_c ]     = 1.*(1.*2.)*fi[ defs.i3_XZ2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = 2.*(1.*2.)*fi[ defs.i3_X2Z2_c ] * 0.25
            fi2[ defs.i3_Y_c ] = 1.*(1.*2.)*fi[ defs.i3_XYZ2_c ] * 0.5
            fi2[ defs.i3_Z_c ] = 1.*(2.*3.)*fi[ defs.i3_XZ3_c  ] * onesixth

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X2Z_c:
        fi2[ defs.i3_F_c ]     = (1.*2.)*1.*fi[ defs.i3_X2Z_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i3_X_c ] = (2.*3.)*1.*fi[ defs.i3_X3Z_c  ] * onesixth
            fi2[ defs.i3_Y_c ] = (1.*2.)*1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_Z_c ] = (1.*2.)*2.*fi[ defs.i3_X2Z2_c ] * 0.25

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XYZ_c:
        fi2[ defs.i3_F_c ]     = 1.*1.*1.*fi[ defs.i3_XYZ_c  ]
        if order >= 4:
            fi2[ defs.i3_X_c ] = 2.*1.*1.*fi[ defs.i3_X2YZ_c ] * 0.5
            fi2[ defs.i3_Y_c ] = 1.*2.*1.*fi[ defs.i3_XY2Z_c ] * 0.5
            fi2[ defs.i3_Z_c ] = 1.*1.*2.*fi[ defs.i3_XYZ2_c ] * 0.5

        return polyeval.general_3D( order-3, fi2, case.xi, case.yi, case.zi, x, out )

    # --------------- 4th order derivatives ---------------
    # if we get here, model order >= 4

    elif diff == defs.i3_X4_c:
        fi2[ defs.i3_F_c ] = 1.*2.*3.*4.*fi[ defs.i3_X4_c ] * one24th
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X3Y_c:
        fi2[ defs.i3_F_c ] = (1.*2.*3.)*1.*fi[ defs.i3_X3Y_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X2Y2_c:
        fi2[ defs.i3_F_c ] = 2.*2.*fi[ defs.i3_X2Y2_c ] * 0.25
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XY3_c:
        fi2[ defs.i3_F_c ] = 1.*(1.*2.*3.)*fi[ defs.i3_XY3_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y4_c:
        fi2[ defs.i3_F_c ] = 1.*2.*3.*4.*fi[ defs.i3_Y4_c ] * one24th
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y3Z_c:
        fi2[ defs.i3_F_c ] = (1.*2.*3.)*1.*fi[ defs.i3_Y3Z_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Y2Z2_c:
        fi2[ defs.i3_F_c ] = 2.*2.*fi[ defs.i3_Y2Z2_c ] * 0.25
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_YZ3_c:
        fi2[ defs.i3_F_c ] = 1.*(1.*2.*3.)*fi[ defs.i3_YZ3_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_Z4_c:
        fi2[ defs.i3_F_c ] = 1.*2.*3.*4.*fi[ defs.i3_Z4_c ] * one24th
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XZ3_c:
        fi2[ defs.i3_F_c ] = 1.*(1.*2.*3.)*fi[ defs.i3_XZ3_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X2Z2_c:
        fi2[ defs.i3_F_c ] = 2.*2.*fi[ defs.i3_X2Z2_c ] * 0.25
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X3Z_c:
        fi2[ defs.i3_F_c ] = (1.*2.*3.)*1.*fi[ defs.i3_X3Z_c ] * onesixth
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_X2YZ_c:
        fi2[ defs.i3_F_c ] = (1.*2.)*1.*1.*fi[ defs.i3_X2YZ_c ] * 0.5
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XY2Z_c:
        fi2[ defs.i3_F_c ] = 1.*(1.*2.)*1.*fi[ defs.i3_XY2Z_c ] * 0.5
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    elif diff == defs.i3_XYZ2_c:
        fi2[ defs.i3_F_c ] = 1.*1.*(1.*2.)*fi[ defs.i3_XYZ2_c ] * 0.5
        return polyeval.general_3D( order-4, fi2, case.xi, case.yi, case.zi, x, out )

    # --------------- end ---------------

    else:
        return -1  # invalid value for "diff"


# case  : infra.Case object with the necessary metadata
#     xi    : in, x0 of the origin (x0,y0) of the fit
#     yi    : in, y0 of the origin (x0,y0) of the fit
#     order : in, degree of the surrogate polynomial
#     fi    : in, fit coefficients
# x     : in, x[k,:] = (x,y) to which to interpolate the model
# out   : out, function values. Must be allocated by caller. Must have as nk elements (one element for each point x).
# diff  : in, one of the i2_* constants in wlsqm.fitter.defs.
#
# Note that the same xi and order must be passed here as for the model fitting.
# fi[] must contain the output the model fitting.
#
cdef int interpolate_2D( infra.Case* case, double[::view.generic,::view.contiguous] x, double* out, int diff ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    cdef int order = case.order
    cdef int k,n

    # Interpolate function value? (the most common case)
    if diff == defs.i2_F_c:
        return polyeval.taylor_2D( order, case.fi, case.xi, case.yi, x, out )

    # Interpolate a derivative of an order higher than the order of the model?
    #
    # The result is identically zero, so we special-case it here.
    #
    # Note that:
    #   - The constants used as values of "diff" are ordered in increasing order of derivatives
    #   - infra.number_of_dofs() returns one-past-end indices for each order of derivatives
    #   - This information is cached into case.no
    #
    elif diff >= case.no:  # infra.number_of_dofs( case.dimension, case.order ):
        n = x.shape[0]
        for k in range(n):
            out[k] = 0.
        return 0  # success

    # Else this derivative may be nonzero, handle it normally.

    # We fill in the coefficient array of the derivative polynomial,
    # shifting the coefficients to their new positions in the array
    # and accounting for exponents that "drop down" in the differentiation.
    #
    # We must also account for the constant factors in the Taylor expansion,
    # since fi[] contains only the DOF value data (and the DOFs are the
    # function value and derivatives at the point xi).
    #
    # Then we use the generic polynomial evaluator.

    cdef double* fi  = case.fi
    cdef double* fi2 = case.fi2  # work space

    # --------------- 1st order derivatives ---------------
    # if we get here, model order >= 1

    if diff == defs.i2_X_c:
        fi2[ defs.i2_F_c ]       = 1.*fi[ defs.i2_X_c    ]
        if order >= 2:
            fi2[ defs.i2_X_c   ] = 2.*fi[ defs.i2_X2_c   ] * 0.5
            fi2[ defs.i2_Y_c   ] = 1.*fi[ defs.i2_XY_c   ]
        if order >= 3:
            fi2[ defs.i2_X2_c  ] = 3.*fi[ defs.i2_X3_c   ] * onesixth
            fi2[ defs.i2_XY_c  ] = 2.*fi[ defs.i2_X2Y_c  ] * 0.5
            fi2[ defs.i2_Y2_c  ] = 1.*fi[ defs.i2_XY2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i2_X3_c  ] = 4.*fi[ defs.i2_X4_c   ] * one24th
            fi2[ defs.i2_X2Y_c ] = 3.*fi[ defs.i2_X3Y_c  ] * onesixth
            fi2[ defs.i2_XY2_c ] = 2.*fi[ defs.i2_X2Y2_c ] * 0.25
            fi2[ defs.i2_Y3_c  ] = 1.*fi[ defs.i2_XY3_c  ] * onesixth

        return polyeval.general_2D( order-1, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_Y_c:
        fi2[ defs.i2_F_c ]       = 1.*fi[ defs.i2_Y_c    ]
        if order >= 2:
            fi2[ defs.i2_X_c   ] = 1.*fi[ defs.i2_XY_c   ]
            fi2[ defs.i2_Y_c   ] = 2.*fi[ defs.i2_Y2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i2_X2_c  ] = 1.*fi[ defs.i2_X2Y_c  ] * 0.5
            fi2[ defs.i2_XY_c  ] = 2.*fi[ defs.i2_XY2_c  ] * 0.5
            fi2[ defs.i2_Y2_c  ] = 3.*fi[ defs.i2_Y3_c   ] * onesixth
        if order >= 4:
            fi2[ defs.i2_X3_c  ] = 1.*fi[ defs.i2_X3Y_c  ] * onesixth
            fi2[ defs.i2_X2Y_c ] = 2.*fi[ defs.i2_X2Y2_c ] * 0.25
            fi2[ defs.i2_XY2_c ] = 3.*fi[ defs.i2_XY3_c  ] * onesixth
            fi2[ defs.i2_Y3_c  ] = 4.*fi[ defs.i2_Y4_c   ] * one24th

        return polyeval.general_2D( order-1, fi2, case.xi, case.yi, x, out )

    # --------------- 2nd order derivatives ---------------
    # if we get here, model order >= 2

    elif diff == defs.i2_X2_c:
        fi2[ defs.i2_F_c ]      = 1.*2.*fi[ defs.i2_X2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i2_X_c  ] = 2.*3.*fi[ defs.i2_X3_c   ] * onesixth
            fi2[ defs.i2_Y_c  ] = 1.*2.*fi[ defs.i2_X2Y_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i2_X2_c ] = 3.*4.*fi[ defs.i2_X4_c   ] * one24th
            fi2[ defs.i2_XY_c ] = 2.*3.*fi[ defs.i2_X3Y_c  ] * onesixth
            fi2[ defs.i2_Y2_c ] = 1.*2.*fi[ defs.i2_X2Y2_c ] * 0.25

        return polyeval.general_2D( order-2, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_XY_c:
        fi2[ defs.i2_F_c ]      = 1.*1.*fi[ defs.i2_XY_c   ]
        if order >= 3:
            fi2[ defs.i2_X_c  ] = 2.*1.*fi[ defs.i2_X2Y_c  ] * 0.5
            fi2[ defs.i2_Y_c  ] = 1.*2.*fi[ defs.i2_XY2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i2_X2_c ] = 3.*1.*fi[ defs.i2_X3Y_c  ] * onesixth
            fi2[ defs.i2_XY_c ] = 2.*2.*fi[ defs.i2_X2Y2_c ] * 0.25
            fi2[ defs.i2_Y2_c ] = 1.*3.*fi[ defs.i2_XY3_c  ] * onesixth

        return polyeval.general_2D( order-2, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_Y2_c:
        fi2[ defs.i2_F_c ]      = 1.*2.*fi[ defs.i2_Y2_c   ] * 0.5
        if order >= 3:
            fi2[ defs.i2_X_c  ] = 1.*2.*fi[ defs.i2_XY2_c  ] * 0.5
            fi2[ defs.i2_Y_c  ] = 2.*3.*fi[ defs.i2_Y3_c   ] * onesixth
        if order >= 4:
            fi2[ defs.i2_X2_c ] = 1.*2.*fi[ defs.i2_X2Y2_c ] * 0.25
            fi2[ defs.i2_XY_c ] = 2.*3.*fi[ defs.i2_XY3_c  ] * onesixth
            fi2[ defs.i2_Y2_c ] = 3.*4.*fi[ defs.i2_Y4_c   ] * one24th

        return polyeval.general_2D( order-2, fi2, case.xi, case.yi, x, out )

    # --------------- 3rd order derivatives ---------------
    # if we get here, model order >= 3

    elif diff == defs.i2_X3_c:
        fi2[ defs.i2_F_c ]     = 1.*2.*3.*fi[ defs.i2_X3_c  ] * onesixth
        if order >= 4:
            fi2[ defs.i2_X_c ] = 2.*3.*4.*fi[ defs.i2_X4_c  ] * one24th
            fi2[ defs.i2_Y_c ] = 1.*2.*3.*fi[ defs.i2_X3Y_c ] * onesixth

        return polyeval.general_2D( order-3, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_X2Y_c:
        fi2[ defs.i2_F_c ]     = (1.*2.)*1.*fi[ defs.i2_X2Y_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i2_X_c ] = (2.*3.)*1.*fi[ defs.i2_X3Y_c  ] * onesixth
            fi2[ defs.i2_Y_c ] = (1.*2.)*2.*fi[ defs.i2_X2Y2_c ] * 0.25

        return polyeval.general_2D( order-3, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_XY2_c:
        fi2[ defs.i2_F_c ]     = 1.*(1.*2.)*fi[ defs.i2_XY2_c  ] * 0.5
        if order >= 4:
            fi2[ defs.i2_X_c ] = 2.*(1.*2.)*fi[ defs.i2_X2Y2_c ] * 0.25
            fi2[ defs.i2_Y_c ] = 1.*(2.*3.)*fi[ defs.i2_XY3_c  ] * onesixth

        return polyeval.general_2D( order-3, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_Y3_c:
        fi2[ defs.i2_F_c ]     = 1.*2.*3.*fi[ defs.i2_Y3_c  ] * onesixth
        if order >= 4:
            fi2[ defs.i2_X_c ] = 1.*2.*3.*fi[ defs.i2_XY3_c ] * onesixth
            fi2[ defs.i2_Y_c ] = 2.*3.*4.*fi[ defs.i2_Y4_c  ] * one24th

        return polyeval.general_2D( order-3, fi2, case.xi, case.yi, x, out )

    # --------------- 4th order derivatives ---------------
    # if we get here, model order >= 4

    elif diff == defs.i2_X4_c:
        fi2[ defs.i2_F_c ] = 1.*2.*3.*4.*fi[ defs.i2_X4_c ] * one24th
        return polyeval.general_2D( order-4, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_X3Y_c:
        fi2[ defs.i2_F_c ] = (1.*2.*3.)*1.*fi[ defs.i2_X3Y_c ] * onesixth
        return polyeval.general_2D( order-4, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_X2Y2_c:
        fi2[ defs.i2_F_c ] = (1.*2.)*(1.*2.)*fi[ defs.i2_X2Y2_c ] * 0.25
        return polyeval.general_2D( order-4, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_XY3_c:
        fi2[ defs.i2_F_c ] = 1.*(1.*2.*3.)*fi[ defs.i2_XY3_c ] * onesixth
        return polyeval.general_2D( order-4, fi2, case.xi, case.yi, x, out )

    elif diff == defs.i2_Y4_c:
        fi2[ defs.i2_F_c ] = 1.*2.*3.*4.*fi[ defs.i2_Y4_c ] * one24th
        return polyeval.general_2D( order-4, fi2, case.xi, case.yi, x, out )

    # --------------- end ---------------

    else:
        return -1  # invalid value for "diff"


# case  : infra.Case object with the necessary metadata
#     xi    : in, the origin (x0) of the fit
#     order : in, degree of the surrogate polynomial
#     fi    : in, fit coefficients
# x     : in, x[k] = (x) to which to interpolate the model
# out   : out, function values. Must be allocated by caller. Must have as nk elements (one element for each point x).
# diff  : in, one of the i1_* constants in wlsqm.fitter.defs.
#
# Note that the same xi and order must be passed here as for the model fitting.
# fi[] must contain the output from the model fitting.
#
cdef int interpolate_1D( infra.Case* case, double[::view.generic] x, double* out, int diff ) nogil:
    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    cdef int order = case.order
    cdef int k,n

    # Interpolate function value? (the most common case)
    if diff == defs.i1_F_c:
        return polyeval.taylor_1D( order, case.fi, case.xi, x, out )

    # Interpolate a derivative higher than order of the model?
    elif diff >= case.no:  # infra.number_of_dofs( case.dimension, case.order ):
        n = x.shape[0]
        for k in range(n):
            out[k] = 0.
        return 0  # success

    # Else this derivative may be nonzero, handle it normally.

    # We fill in the coefficient array of the derivative polynomial,
    # shifting the coefficients to their new positions in the array
    # and accounting for exponents that "drop down" in the differentiation.
    #
    # We must also account for the constant factors in the Taylor expansion,
    # since fi[] contains only the DOF value data (and the DOFs are the
    # function value and derivatives at the point xi).
    #
    # Then we use the generic polynomial evaluator.

    cdef double* fi  = case.fi
    cdef double* fi2 = case.fi2  # work space

    # --------------- 1st order derivatives ---------------
    # if we get here, model order >= 1

    if diff == defs.i1_X_c:
        fi2[ defs.i1_F_c ]      = 1.*fi[ defs.i1_X_c  ]
        if order >= 2:
            fi2[ defs.i1_X_c  ] = 2.*fi[ defs.i1_X2_c ] * 0.5
        if order >= 3:
            fi2[ defs.i1_X2_c ] = 3.*fi[ defs.i1_X3_c ] * onesixth
        if order >= 4:
            fi2[ defs.i1_X3_c ] = 4.*fi[ defs.i1_X4_c ] * one24th

        return polyeval.general_1D( order-1, fi2, case.xi, x, out )

    # --------------- 2nd order derivatives ---------------
    # if we get here, model order >= 2

    elif diff == defs.i1_X2_c:
        fi2[ defs.i1_F_c ]      = 1.*2.*fi[ defs.i1_X2_c ] * 0.5
        if order >= 3:
            fi2[ defs.i1_X_c  ] = 2.*3.*fi[ defs.i1_X3_c ] * onesixth
        if order >= 4:
            fi2[ defs.i1_X2_c ] = 3.*4.*fi[ defs.i1_X4_c ] * one24th

        return polyeval.general_1D( order-2, fi2, case.xi, x, out )

    # --------------- 3rd order derivatives ---------------
    # if we get here, model order >= 3

    elif diff == defs.i1_X3_c:
        fi2[ defs.i1_F_c ]     = 1.*2.*3.*fi[ defs.i1_X3_c ] * onesixth
        if order >= 4:
            fi2[ defs.i1_X_c ] = 2.*3.*4.*fi[ defs.i1_X4_c ] * one24th

        return polyeval.general_1D( order-3, fi2, case.xi, x, out )

    # --------------- 4th order derivatives ---------------
    # if we get here, model order >= 4

    elif diff == defs.i1_X4_c:
        fi2[ defs.i1_F_c ] = 1.*2.*3.*4.*fi[ defs.i1_X4_c ] * one24th
        return polyeval.general_1D( order-4, fi2, case.xi, x, out )

    # --------------- end ---------------

    else:
        return -1  # invalid value for "diff"




# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# Centralized memory allocation infrastructure.
#
# JJ 2016-11-30

# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True

# Total memory needed for arrays:
#    - no*sizeof(int) bytes for o2r; one shared copy is enough (bypass custom alloc, since only remap(), which needs this, computes "nr")
#    - no*sizeof(int) bytes for r2o; one shared copy is enough (                                --''--                                  )
#    - nprob*nk*no*sizeof(double) bytes for c   (actually, sum(nk_j, j in problems)*no*sizeof(double))
#    - nprob*nr*sizeof(double) bytes for row_scale
#    - nprob*nr*sizeof(double) bytes for column_scale
#    - nprob*nr*nr*sizeof(double) bytes for A
#    - nprob*nr*sizeof(int) bytes for ipiv
#    - solve:
#    -   if do_sens, ntasks*nr*(nk+1)*sizeof(double) bytes for wrk
#           - use max(nk_j) here to fit the largest problem instance (since any thread may run it)
#    -   if not do_sens, ntasks*nr*sizeof(double) bytes for wrk
#    - iterative refinement:
#      - ntasks*nk*sizeof(double) bytes for fk_tmp
#           - max(nk_j) here too, same reason
#      - ntasks*no*sizeof(double) bytes for fi_tmp

from __future__ import division
from __future__ import absolute_import

from libc.stdlib cimport malloc, free
from libc.math cimport sqrt

cimport wlsqm.fitter.defs as defs  # C constants

# use GCC's intrinsics for counting the number of set bits in an int
#
# See
#   http://stackoverflow.com/questions/109023/how-to-count-the-number-of-set-bits-in-a-32-bit-integer (algorithms, suggestions)
#   https://gist.github.com/craffel/e470421958cad33df550 (Cython defs; on popcounting a NumPy array)
#
cdef extern int __builtin_popcount(unsigned int) nogil
cdef extern int __builtin_popcountll(unsigned long long) nogil

#####################################
# Helper functions
#####################################

# Return number of DOFs in the original (unreduced) system.
#
# dimension : in, number of space dimensions (1, 2 or 3)
# order     : in, the order of the polynomial to be fitted
#
cdef int number_of_dofs( int dimension, int order ) nogil:
    if dimension not in [1,2,3]:
        return -1
#        with gil:
#            raise ValueError( "dimension must be 1, 2 or 3; got %d" % dimension )
    if order not in [0,1,2,3,4]:
        return -2
#        with gil:
#            raise ValueError( "order must be 0, 1, 2, 3 or 4; got %d" % order )

    cdef int no
    if dimension == 3:
        if order == 4:
            no = defs.i3_4th_end_c
        elif order == 3:
            no = defs.i3_3rd_end_c
        elif order == 2:
            no = defs.i3_2nd_end_c
        elif order == 1:
            no = defs.i3_1st_end_c
        else: # order == 0:
            no = defs.i3_0th_end_c
    elif dimension == 2:
        if order == 4:
            no = defs.i2_4th_end_c
        elif order == 3:
            no = defs.i2_3rd_end_c
        elif order == 2:
            no = defs.i2_2nd_end_c
        elif order == 1:
            no = defs.i2_1st_end_c
        else: # order == 0:
            no = defs.i2_0th_end_c
    else: # dimension == 1:
        if order == 4:
            no = defs.i1_4th_end_c
        elif order == 3:
            no = defs.i1_3rd_end_c
        elif order == 2:
            no = defs.i1_2nd_end_c
        elif order == 1:
            no = defs.i1_1st_end_c
        else: # order == 0:
            no = defs.i1_0th_end_c

    return no

# Return the number of DOFs in the reduced system, corresponding to an original (unreduced) number of DOFs  n  and a knowns mask.
#
# n    : in, number of DOFs in the original (unreduced) system
# mask : in, bitmask of knowns
#
cdef int number_of_reduced_dofs( int n, long long mask ) nogil:
    cdef int ne = __builtin_popcountll(mask)  # number of eliminated DOFs = number of bits set in mask
    return n - ne  # remaining DOFs

# Reduce the system size by removing the rows/columns for knowns from the DOF numbering.
#
# Specifically:
#
# Given a bitmask of DOFs to eliminate, construct DOF number mappings
# between the original full equation system and the reduced equation system.
#
# o2r  : out, mapping original --> reduced;  size (n,), must be allocated by caller
# r2o  : out, mapping reduces  --> original; size (n,), must be allocated by caller
# n    : in, number of DOFs in the original (unreduced) system
# mask : in, bitmask of knowns
#
# return value: the number of DOFs in the reduced system.
#
# In the arrays, non-existent DOFs will be represented by the special value -1.
#
# In o2r (original->reduced), non-existent DOFs are those that were eliminated
# (hence have no index in the reduced system).
#
# In r2o (reduced->original), non-existent DOFs are those with  index >= n_reduced,
# where  n_reduced = (n - n_eliminated),  since the reduced system has only n_reduced DOFs in total.
#
cdef int remap( int* o2r, int* r2o, int n, long long mask ) nogil:  # o = original, r = reduced
    # We always start the elimination with a full range(n) of DOFs.
    #
    # For example, if we have 4 DOFs, and we would like to eliminate the DOF "1",
    # we construct the following mappings:
    #
    #    orig -> reduced
    #
    #       0 ->  0
    #       1 -> -1 (original DOF "1" does not exist in reduced system)
    #       2 ->  1
    #       3 ->  2
    #
    # reduced -> orig
    #
    #       0 ->  0
    #       1 ->  2
    #       2 ->  3
    #       3 -> -1 (in the reduced system, there is no DOF "3")
    #
    # The right-hand sides can be expressed in array form, using the left-hand side as the array index:
    #
    # orig->reduced: [0, -1, 1, 2]
    # reduced->orig: [0, 2, 3, -1]
    #
    # These arrays are the output format.
    #
    # This says that e.g. the DOF "2" of orig maps to DOF "1" of reduced (array orig->reduced, index 2).
    # The DOF "1" of reduced maps to DOF "2" of orig (array reduced->orig, index 1).

    # We first generate orig -> reduced.
    #
    cdef int j, k=0  # k = first currently available DOF number (0-based) in the reduced system
    for j in range(n):
        if mask & (1LL << j):  # eliminate this DOF?
            o2r[j] = -1
        else:
            o2r[j] = k
            k += 1  # a DOF was introduced into the reduced system

    # k is now the number of DOFs in the reduced system

    # Construct the inverse to obtain reduced -> orig. See the example above.
    #
    for j in range(n):
        if o2r[j] == -1:
            continue
        r2o[ o2r[j] ] = j

    # In the reduced -> orig mapping, set the rest to -1, since the reduced system
    # has less DOFs than the original one.
    #
    for j in range(k, n):
        r2o[j] = -1

    return k


#################################################
# class Allocator:
#################################################

# To avoid memory fragmentation in the case with many instances of the model being fitted at once,
# we use a custom memory allocator.
#
# This is very simplistic; we do not need to support the re-use of already allocated blocks.
#
# Example:
#    int total_size_bytes = 1000000  # 1 MB
#    Allocator* a = Allocator_new( ALLOC_MODE_ONEBIGBUFFER, total_size_bytes )
#    int* my_block_1 = <int*>Allocator_malloc( a, 100*sizeof(int) )
#    # ...other Allocator_malloc()'s...
#    # ...
#    Allocator_free( a, my_block_1 )
#    # ...other Allocator_free()'s...
#    Allocator_del( a )

# Object-oriented programming, C style.
cdef int ALLOC_MODE_PASSTHROUGH  = 1  # pass each call through to C malloc/free
cdef int ALLOC_MODE_ONEBIGBUFFER = 2  # pre-allocate one big buffer to fit everything in

# Constructor.
#
# This class is only intended to be used from a Python thread.
#
# Note that ALLOC_MODE_PASSTRHOUGH doesn't use total_size_bytes, but .pxd files do not support default values for function arguments,
# and Cython's "int total_size_bytes=*" syntax (in the .pxd file) does not support nogil functions.
#
cdef Allocator* Allocator_new( int mode, int total_size_bytes ) nogil except <Allocator*>0:
    cdef Allocator* self = <Allocator*>malloc( sizeof(Allocator) )
    if self == <Allocator*>0:  # we promised Cython not to return NULL, so we must raise if the malloc fails
        with gil:
            raise MemoryError("Out of memory trying to allocate an Allocator object")

    if mode == ALLOC_MODE_ONEBIGBUFFER  and  total_size_bytes > 0:
        self.buffer = malloc( total_size_bytes )
        if self.buffer == <void*>0:
            with gil:
                raise MemoryError("Out of memory trying to allocate a buffer of %d bytes" % (total_size_bytes))
    else:
        self.buffer = <void*>0

    self.mode       = mode
    self.size_total = total_size_bytes
    self.p          = self.buffer
    self.size_used  = 0

    return self

cdef void* Allocator_malloc( Allocator* self, int size_bytes ) nogil:
    if self.mode == ALLOC_MODE_PASSTHROUGH:
#        with gil:
#            print "directly allocating %d bytes" % (size_bytes)
        return malloc( size_bytes )

    # else...

    # pathological case: no buffer, can't allocate
    if self.buffer == <void*>0:
        return <void*>0

    # check that there is enough space remaining in the buffer
    cdef int size_remaining = self.size_total - self.size_used
    if size_bytes > size_remaining:
#        with gil:
#            print "buffer full, cannot allocate %d bytes" % (size_bytes)  # DEBUG
        return <void*>0

    cdef void* p
    with gil:  # since we are called from Python threads only, we can use the GIL to make the operation thread-safe. (TODO/FIXME: well, not exactly, see e.g. http://www.slideshare.net/dabeaz/an-introduction-to-python-concurrency )
#        print "reserving %d bytes from buffer of size %d; after alloc, %d bytes remaining" % (size_bytes, self.size_total, size_remaining - size_bytes)  # DEBUG
        p = self.p
        self.p = <void*>( (<char*>p) + size_bytes )
        self.size_used += size_bytes

    return p

cdef void Allocator_free( Allocator* self, void* p ) nogil:
    if self.mode == ALLOC_MODE_PASSTHROUGH:
        free( p )
    # else do nothing; this simplistic allocator doesn't reuse blocks once they are allocated

cdef int Allocator_size_remaining( Allocator* self ) nogil:
    return self.size_total - self.size_used

# Destructor.
cdef void Allocator_del( Allocator* self ) nogil:
    if self != <Allocator*>0:
        free( self.buffer )
        free( self )


#################################################
# class CaseManager:
#################################################

# Constructor.
#
# max_cases is mandatory (with an invalid default value since no sane default can exist).
#
# ntasks is for parallel processing at solve time; effectively, it specifies how many per-task arrays to allocate.
# When processing serially, use the value 1.
#
cdef CaseManager* CaseManager_new( int dimension, int do_sens, int iterative, int max_cases, int ntasks ) nogil except <CaseManager*>0:
    # Generally speaking, fixing the array size at instantiation time is stupid (an automatically expanding buffer would be better),
    # but considering that this class has only one actual user, where we do know max_cases in advance, it is fine for our purposes.
    if max_cases < 1:
        with gil:
            raise ValueError("Must specify max_cases > 0 when creating a CaseManager.")

    cdef CaseManager* self = <CaseManager*>malloc( sizeof(CaseManager) )
    if self == <CaseManager*>0:  # we promised Cython not to return NULL, so we must raise if the malloc fails
        with gil:
            raise MemoryError("Out of memory trying to allocate an CaseManager object")

    # init parallel proc
    #
    self.ntasks  = ntasks
    self.wrks    = <double**>malloc( ntasks*sizeof(double*) )
    self.fk_tmps = <double**>malloc( ntasks*sizeof(double*) )
    self.fi_tmps = <double**>malloc( ntasks*sizeof(double*) )
    for j in range(ntasks):  # init to NULL needed to gracefully handle errors in CaseManager_allocate()
        self.wrks[j]    = <double*>0
        self.fk_tmps[j] = <double*>0
        self.fi_tmps[j] = <double*>0

    # init storage for Case pointers
    #
    self.cases = <Case**>malloc( max_cases*sizeof(Case*) )
    self.max_cases = max_cases
    self.ncases = 0

    # save metadata
    #
    self.dimension = dimension
    self.do_sens   = do_sens
    self.iterative = iterative

    # these will be set up at allocate time
    #
    self.mal          = <Allocator*>0
    self.bytes_needed = 0

    return self

# Add a case to this manager.
#
# This means the manager will manage the memory for the Case, and at destruction time,
# will also destroy the managed Case.
#
# Up to max_cases Case objects can be added to the manager (see CaseManager_new()).
#
# Case_new() will call this automatically, if a manager is specified.
#
cdef int CaseManager_add( CaseManager* self, Case* case ) nogil except -1:
    # sanity check remaining space
    if self.ncases == self.max_cases:
        with gil:
            raise MemoryError("Case pointer buffer full, max_cases = %d reached" % self.max_cases)

    # sanity check Case metadata for compatibility with this CaseManager instance
    if case.dimension != self.dimension:
        with gil:
            raise ValueError("Cannot add case with different dimension = %d; this manager has dimension = %d" % (case.dimension, self.dimension))
    if case.do_sens != self.do_sens:
        with gil:
            raise ValueError("Cannot add case with different setting for do_sens = %d; this manager has do_sens = %d" % (case.do_sens, self.do_sens))
    if case.iterative != self.iterative:
        with gil:
            raise ValueError("Cannot add case with different setting for iterative = %d; this manager has iterative = %d" % (case.iterative, self.iterative))

    # add the Case to the managed cases.
    self.cases[self.ncases] = case
    self.ncases += 1

    return 0

# Finish adding Case objects. Prepare the manaager for solving.
#
# This should be called exactly once (per instance of CaseManager), after all cases have been CaseManager_add()'d.
#
cdef int CaseManager_commit( CaseManager* self ) nogil except -1:
    return CaseManager_allocate( self )

# Create the memory buffer that will contain the data arrays for all cases.
#
cdef int CaseManager_allocate( CaseManager* self ) nogil except -1:
    cdef int j, problem_instance_bytes=0, task_bytes=0
    cdef int max_nk=0, max_no=0, max_nr=0
    cdef int size_wrk=0, size_fk_tmp=0, size_fi_tmp=0
    cdef BufferSizes sizes
    cdef Case* case
    with gil:
        try:
            if self.ncases == 0:
                raise ValueError("No cases; add some before allocating")

            # Determine total amount of memory needed by the per-problem-instance arrays.
            #
            # Also, find max_nk for allocation of per-task arrays. (nk may vary across cases)
            #
            for j in range(self.ncases):
                case = self.cases[j]
                Case_determine_sizes( case, &sizes )

                problem_instance_bytes += sizes.total

                if case.nk > max_nk:
                    max_nk = case.nk
                if case.no > max_no:
                    max_no = case.no
                if case.nr > max_nr:
                    max_nr = case.nr

            # Determine maximum needed size for one instance of the per-task arrays,
            # when working on this set of cases.
            #
            if self.do_sens:
                size_wrk = max_nr*(max_nk + 1)*sizeof(double)
            else:
                size_wrk = max_nr*sizeof(double)

            if self.iterative:
                size_fk_tmp = max_nk*sizeof(double)
                size_fi_tmp = max_no*sizeof(double)
            else:
                size_fk_tmp = 0
                size_fi_tmp = 0

            # The total for the per-task arrays is then just ntasks copies:
            #
            task_bytes = self.ntasks*(size_wrk + size_fk_tmp + size_fi_tmp)

            # Final total of memory needed is thus:
            #
            self.bytes_needed = task_bytes + problem_instance_bytes

            # NOTE: this may be big (e.g. ~5.5kB per problem instance for dimension=2, order=4, nk=25,
            #                        so for a moderate number of 1e4 problem instances, this is already 55MB)
            #
            # The good news is that even if multiple fits (against different data) are performed with the same points,
            # we can simply let the buffer be - there is no need to re-create it for each run.
            #
            self.mal = Allocator_new( mode=ALLOC_MODE_ONEBIGBUFFER, total_size_bytes=self.bytes_needed )

            # Allocate the per-task arrays.
            #
            for j in range(self.ntasks):
                self.wrks[j]    = <double*>Allocator_malloc( self.mal, size_wrk )
                self.fk_tmps[j] = <double*>Allocator_malloc( self.mal, size_fk_tmp )
                self.fi_tmps[j] = <double*>Allocator_malloc( self.mal, size_fi_tmp )

            # Finally, tell the Case objects to allocate their memory.
            #
            # They will automatically grab our allocator, since they are in managed mode.
            #
            for j in range(self.ncases):
                Case_allocate( self.cases[j] )

        except:
            # on error, leave the CaseManager in the state it was in before this method was called.
            CaseManager_deallocate( self )
            raise

    return 0

# The opposite of CaseManager_allocate().
#
cdef void CaseManager_deallocate( CaseManager* self ) nogil:
    if self != <CaseManager*>0:
        self.bytes_needed = 0

        if self.mal != <Allocator*>0:  # the allocator instantiation may have failed, so make sure we have an allocator before attempting this
            # the managed Case objects also use our allocator
            for j in range(self.ncases):
                Case_deallocate( self.cases[j] )  # it is safe to Case_deallocate() also a Case that has not yet been Case_allocate()'d.

            for j in range(self.ntasks):
                Allocator_free( self.mal, self.fi_tmps[j] )
                self.fi_tmps[j] = <double*>0
                Allocator_free( self.mal, self.fk_tmps[j] )
                self.fk_tmps[j] = <double*>0
                Allocator_free( self.mal, self.wrks[j] )
                self.wrks[j]    = <double*>0

            Allocator_del( self.mal )
            self.mal = <Allocator*>0

# Destructor. Destroys also the managed Case objects.
#
cdef void CaseManager_del( CaseManager* self ) nogil:
    cdef int j
    if self != <CaseManager*>0:
        CaseManager_deallocate( self )

        # destroy the managed Case objects
        for j in range(self.ncases):
            Case_del( self.cases[j] )

        # free manually allocated storage
        free( self.cases )
        free( self.fi_tmps )
        free( self.fk_tmps )
        free( self.wrks )

        free( self )


#################################################
# class Case:
#################################################

# Constructor.
#
# This class is only intended to be instantiated from a Python thread.
#
# manager: an already existing CaseManager object to use, to share the memory allocator among a set of cases.
#          The cases must have the same dimension, do_sens, iterative.
#
#          If null, an Allocator will be created locally.
#
# host:    for guest mode, an existing Case object to use. The geometry data (o2r,r2o,c,w,A,row_scale,col_scale,ipiv) will be borrowed off the host,
#          and no local copies will be created.
#
#          This can be used to save both memory and time when different fields (in an IBVP problem) live on the exact same geometry.
#          "Geometry" includes both xi,yi.zi (point "xi") and the neighbor set (points "xk"; see wlsqm.fitter.impl.make_c_?D()).
#
#          Thus, the host Case instance must have the exact same parameters (and geometry!) as the Case instance being created.
#          "Parameters" include  dimension, order, nk, knowns, weighting_method.
#
#          The parameter match is not checked! See wlswm2_expert.ExpertSolver for correct usage.
#          (It does some rudimentary checking, but does not check the geometry.)
#
#          When using guest mode, the calling code must make sure the host instance stays alive at least as long as its guest instances,
#          or hope for a crash.
#
#          If null, the geometry data will be allocated locally.
#
cdef Case* Case_new( int dimension, int order, double xi, double yi, double zi, int nk, long long knowns, int weighting_method, int do_sens, int iterative, CaseManager* manager, Case* host ) nogil except <Case*>0:
    cdef Case* self = <Case*>malloc( sizeof(Case) )
    if self == <Case*>0:  # we promised Cython not to return NULL, so we must raise if the malloc fails
        with gil:
            raise MemoryError("Out of memory trying to allocate a Case object")

    # tag unused components as NaN
    cdef double nan = 0./0.  # NaN as per IEEE-754
    self.xi = xi
    self.yi = yi  if dimension >= 2  else  nan
    self.zi = zi  if dimension == 3  else  nan

    # init data pointers to NULL to make it safe to dealloc partially initialized Case (when something goes wrong)
    self.o2r       = <int*>0
    self.r2o       = <int*>0
    self.c         = <double*>0
    self.w         = <double*>0
    self.A         = <double*>0
    self.row_scale = <double*>0
    self.col_scale = <double*>0
    self.ipiv      = <int*>0
    self.fi        = <double*>0
    self.fi2       = <double*>0
    self.wrk       = <double*>0
    self.fk_tmp    = <double*>0
    self.fi_tmp    = <double*>0

    if host == <Case*>0:
        self.geometry_owned = 1

        # set condition numbers to nan until computed (only computed if wlsqm.fitter.impl.prepare() is called with the debug flag set!)
        self.cond_orig   = nan
        self.cond_scaled = nan

    else:
        self.geometry_owned = 0
        self.o2r            = host.o2r
        self.r2o            = host.r2o
        self.c              = host.c
        self.w              = host.w
        self.A              = host.A
        self.row_scale      = host.row_scale
        self.col_scale      = host.col_scale
        self.ipiv           = host.ipiv

        # these may have been computed by host (this only works if the host has had preprocess_A() called on it already, but this is the best we can do)
        self.cond_orig      = host.cond_orig
        self.cond_scaled    = host.cond_scaled

    # Use the data from CaseManager if given
    self.have_manager = (manager != <CaseManager*>0)
    self.manager      = manager        # (copies the pointer also if NULL)
    self.mal          = <Allocator*>0  # this will be filled at allocate time

    # determine number of DOFs in the original (unreduced) and reduced systems (needed to determine array sizes)
    cdef int no = number_of_dofs( dimension, order )
    cdef int nr = number_of_reduced_dofs( no, knowns )

    # save metadata
    self.dimension        = dimension
    self.order            = order
    self.knowns           = knowns
    self.weighting_method = weighting_method
    self.no               = no
    self.nr               = nr
    self.nk               = nk
    self.do_sens          = do_sens
    self.iterative        = iterative

    # Now the Case is in a half-initialized state, with metadata available, but no memory allocated yet.

    if self.have_manager:
        # In managed mode, cases automatically add themselves to the manager.
        with gil:
            try:
                CaseManager_add( self.manager, self )  # this may raise if the buffer is full
            except:
                free( self )
                raise
    else:
        # In unmanaged mode, for caller convenience, cases fully initialize themselves, since there is no separate allocate step
        # that depends on having all the cases available (to compute the final buffer size).
        Case_allocate( self )

    return self

# Getters for work space pointers for parallel task "taskid" (0, 1, ..., ntasks-1).
#
# In managed mode, the work spaces live in the manager (there are ntasks copies).
#
# In unmanaged mode, each Case has its own work space.
#
cdef double* Case_get_wrk( Case* self, int taskid ) nogil:
    if self.have_manager:
        return self.manager.wrks[taskid]
    else:
        return self.wrk

cdef double* Case_get_fk_tmp( Case* self, int taskid ) nogil:
    if self.have_manager:
        return self.manager.fk_tmps[taskid]
    else:
        return self.fk_tmp

cdef double* Case_get_fi_tmp( Case* self, int taskid ) nogil:
    if self.have_manager:
        return self.manager.fi_tmps[taskid]
    else:
        return self.fi_tmp

# Helper: convert an (nk,) array of squared distances to corresponding weights.
#
# w      : in/out. On entry, squared distances from xi to each xk.
#                  On exit, weight factors for each xk.
# nk     : in, number of neighbor points (i.e. points xk)
# max_d2 : in, the largest squared distance seen (i.e. the max element of input w).
#          This is used for normalization.
# weighting_method : in, one of the constants WEIGHT_*. Specifies the type of weighting to use;
#                    different weightings are good for different use cases of WLSQM.
#
cdef void Case_make_weights( Case* self, double max_d2 ) nogil:
    cdef double* w = self.w
    cdef int nk    = self.nk
    cdef int weighting_method = self.weighting_method

    # no-op in guest mode (weights already computed in the host Case instance)
    if not self.geometry_owned:
        return

    cdef int k
    cdef double d2, tmp
    if weighting_method == defs.WEIGHT_UNIFORM_c:
        # Trivial weighting. Don't use distance information, treat all points as equally important.
        #
        # This gives the best overall fit of function values across all points xk,
        # at the cost of accuracy of derivatives at the point xi.
        #
        # (Essentially, this cost is because derivatives are local, so the information
        #  from far-away points corrupts them.)
        #
        for k in range(nk):
            w[k] = 1.

    else: # weighting_method == defs.WEIGHT_CENTER_c:
        # Emphasize points close to xi.
        #
        # Improves the fit of derivatives at the point xi, at the cost of the overall fit
        # of function values at points xk that are (relatively speaking) distant from xi.
        #
        for k in range(nk):
            d2 = w[k]  # the array w originally contains squared distances (without normalization)

            # distance squared, flipped on the distance axis (fast falloff near origin)
            DEF alpha = 1e-4  # weight remaining at maximum distance
            DEF beta  = 1. - alpha
            tmp = 1. - sqrt(d2 / max_d2)
            w[k] = alpha + beta * tmp*tmp

# Determine how many bytes of memory this Case will need for storing its arrays.
#
# Write the result into the given BufferSizes struct.
#
cdef void Case_determine_sizes( Case* self, BufferSizes* sizes ) nogil:
    cdef int no        = self.no
    cdef int nr        = self.nr
    cdef int nk        = self.nk
    cdef int do_sens   = self.do_sens
    cdef int iterative = self.iterative

    if self.geometry_owned:
        sizes.o2r       = no*sizeof(int)        # (no,)
        sizes.r2o       = no*sizeof(int)        # (no,)
        sizes.c         = nk*no*sizeof(double)  # (nk,no), C-contiguous
        sizes.w         = nk*sizeof(double)     # (nk,)
        sizes.A         = nr*nr*sizeof(double)  # (nr, nr), Fortran-contiguous
        sizes.row_scale = nr*sizeof(double)     # (nr,)
        sizes.col_scale = nr*sizeof(double)     # (nr,)
        sizes.ipiv      = nr*sizeof(int)        # (nr,)
    else:
        # This function computes only bytes needed, so in guest mode we can put zeroes here.
        # This function is not used for determining the number of elements in anything.
        sizes.o2r       = 0
        sizes.r2o       = 0
        sizes.c         = 0
        sizes.w         = 0
        sizes.A         = 0
        sizes.row_scale = 0
        sizes.col_scale = 0
        sizes.ipiv      = 0

    # The coefficient array is always needed.
    #
    sizes.fi        = no*sizeof(double)     # (no,)

    # For any polynomial of degree  d >= 1,  its (non-zero) derivatives are a polynomial of degree  d - 1.
    # In the zeroth order case, the derivative is everywhere zero.
    cdef int no2    = number_of_dofs( self.dimension, self.order - 1 )  if self.order >= 1  else  0
    sizes.fi2       = no2*sizeof(double)    # (no2,)

    # per-task work space arrays
    #
    if self.have_manager:
        # in managed mode, CaseManager will allocate one copy of the per-task (not per-problem-instance) arrays
        sizes.wrk    = 0
        sizes.fk_tmp = 0
        sizes.fi_tmp = 0
    else:
        # unmanaged mode - allocate also the per-task arrays locally.
        # see the header comment of solve() for the solver work space sizes
        if do_sens:
            sizes.wrk = nr*(nk + 1)*sizeof(double)
        else:
            sizes.wrk = nr*sizeof(double)

        if iterative:
            sizes.fk_tmp = nk*sizeof(double)
            sizes.fi_tmp = no*sizeof(double)
        else:
            sizes.fk_tmp = 0
            sizes.fi_tmp = 0

    sizes.total = sizes.o2r + sizes.r2o \
                + sizes.c + sizes.w \
                + sizes.A + sizes.row_scale + sizes.col_scale + sizes.ipiv \
                + sizes.fi + sizes.fi2 \
                + sizes.wrk \
                + sizes.fk_tmp + sizes.fi_tmp

# Load user-given data into the coefficients fi.
#
# The length of the input is assumed to be self.no.
#
# This can be used to populate knowns.
#
cdef void Case_set_fi( Case* self, double* fi ) nogil:
    cdef double* my_fi = self.fi
    cdef int no        = self.no
    cdef int om
    for om in range(no):
        my_fi[om] = fi[om]

# Populate user-given array of length self.no
# with the solution data (coefficients self.fi).
#
cdef void Case_get_fi( Case* self, double* out ) nogil:
    cdef double* my_fi = self.fi
    cdef int no        = self.no
    cdef int om
    for om in range(no):
        out[om] = my_fi[om]

# Perform memory allocation.
#
cdef int Case_allocate( Case* self ) nogil except -1:
    # At this point, the constructor has finished, so we have no, nr, nk and the flags (do_sens, iterative).
    #
    # Calculate the (space-)optimal buffer size for ONEBIGBUFFER mode.
    #
    # (This also calculates the various individual sizes, which we will use to actually allocate the memory.)
    #
    cdef BufferSizes sizes
    Case_determine_sizes( self, &sizes )

    # Acquire or create the custom memory allocator to allocate storage for actual data.
    #
    cdef int size_remaining=-1
    cdef Allocator* mal
    with gil:
        try:
            if self.have_manager:  # managed mode - external allocator given; check that it has enough space for us.
                mal = self.manager.mal
                size_remaining = Allocator_size_remaining( mal )
                if size_remaining < sizes.total:
                    raise MemoryError("%d bytes of memory needed, but the given allocator has only %d bytes remaining." % (sizes.total, size_remaining))

            else:  # unmanaged mode - instantiate our own Allocator. The Allocator constructor will raise MemoryError if it runs out of memory.
                mal = Allocator_new( mode=ALLOC_MODE_ONEBIGBUFFER, total_size_bytes=sizes.total )
        except:
            free( self )
            raise
    self.mal = mal

    # Allocate the storage, using the custom allocator.
    #
    if self.geometry_owned:
        self.o2r       = <int*>   Allocator_malloc( mal, sizes.o2r )
        self.r2o       = <int*>   Allocator_malloc( mal, sizes.r2o )

        self.c         = <double*>Allocator_malloc( mal, sizes.c )
        self.w         = <double*>Allocator_malloc( mal, sizes.w )

        self.A         = <double*>Allocator_malloc( mal, sizes.A )
        self.row_scale = <double*>Allocator_malloc( mal, sizes.row_scale )
        self.col_scale = <double*>Allocator_malloc( mal, sizes.col_scale )
        self.ipiv      = <int*>   Allocator_malloc( mal, sizes.ipiv )

    self.fi        = <double*>Allocator_malloc( mal, sizes.fi )

    if sizes.fi2:
        self.fi2   = <double*>Allocator_malloc( mal, sizes.fi2 )
    else:
        self.fi2   = <double*>0

    if self.have_manager:
        # in managed mode, CaseManager will allocate one copy of the per-task (not per-problem-instance) arrays
        self.wrk    = <double*>0
        self.fk_tmp = <double*>0
        self.fi_tmp = <double*>0

    else:
        # unmanaged mode - allocate the per-task arrays locally.
        self.wrk        = <double*>Allocator_malloc( mal, sizes.wrk )

        if self.iterative:
            self.fk_tmp = <double*>Allocator_malloc( mal, sizes.fk_tmp )
            self.fi_tmp = <double*>Allocator_malloc( mal, sizes.fi_tmp )
        else:
            self.fk_tmp = <double*>0
            self.fi_tmp = <double*>0

    # memory allocated; populate o2r and r2o
    #
    # (in guest mode, this has already been done in the host Case instance)
    #
    if self.geometry_owned:
        remap( self.o2r, self.r2o, self.no, self.knowns )

    return 0

# The opposite of Case_allocate().
#
cdef void Case_deallocate( Case* self ) nogil:
    if self != <Case*>0:
        # No guarantee that we'll be called only from the destructor;
        # we must set any deallocated pointers to NULL.
        Allocator_free( self.mal, self.fi_tmp )
        self.fi_tmp = <double*>0
        Allocator_free( self.mal, self.fk_tmp )
        self.fk_tmp = <double*>0

        Allocator_free( self.mal, self.wrk )
        self.wrk = <double*>0

        Allocator_free( self.mal, self.fi2 )
        self.fi2 = <double*>0
        Allocator_free( self.mal, self.fi )
        self.fi = <double*>0

        if self.geometry_owned:
            Allocator_free( self.mal, self.ipiv )
            self.ipiv = <int*>0
            Allocator_free( self.mal, self.col_scale )
            self.col_scale = <double*>0
            Allocator_free( self.mal, self.row_scale )
            self.row_scale = <double*>0
            Allocator_free( self.mal, self.A )
            self.A = <double*>0

            Allocator_free( self.mal, self.w )
            self.w = <double*>0
            Allocator_free( self.mal, self.c )
            self.c = <double*>0

            Allocator_free( self.mal, self.r2o )
            self.r2o = <int*>0
            Allocator_free( self.mal, self.o2r )
            self.o2r = <int*>0
        else:
            # In guest mode, the allocation of these arrays is managed by the host Case instance.
            self.ipiv      = <int*>0
            self.col_scale = <double*>0
            self.row_scale = <double*>0
            self.A         = <double*>0
            self.w         = <double*>0
            self.c         = <double*>0
            self.r2o       = <int*>0
            self.o2r       = <int*>0

# Destructor.
#
cdef void Case_del( Case* self ) nogil:
    if self != <Case*>0:
        Case_deallocate( self )

        if not self.have_manager:
            Allocator_del( self.mal )

        free( self )




# -*- coding: utf-8 -*-
#
# Cython interface for lapackdrivers.pyx.
#
# Naming scheme (in shellglob notation):
#   *s = multiple RHS (but with the same LHS for all).
#        These are one-shot deals that reuse the matrix factorization internally.
#        However, the pivot information is not returned, so the matrix A
#        is destroyed (overwritten) during the call.
#
#   m* = multiple LHS (a separate single RHS for each)
#        These simply loop over the problem instances.
#
#   *p = parallel (multi-threaded using OpenMP)
#        These introduce parallel looping over problem instances.
#
#   *factor*   = the routine that factors the matrix and generates pivot information.
#   *factored* = the solver routine that uses the factored matrix and pivot information.
#        These are useful for solving with many RHSs, when all the RHSs
#        are not available at once (e.g. in PDE solvers, timestepping
#        with a mass matrix that remains constant in time).
#
#   *_c = C version without memoryviews (only visible from Cython). Can be more convenient
#         for use in nogil blocks, in cases where the arrays need to be allocated dynamically (with malloc).
#         The purpose of the C version is to avoid the need to acquire the GIL to create a memoryview
#         into a malloc()'d array.
#
#         This .pxd file offers access to only the C versions. To import the corresponding Python versions
#         of the routines (same name, without the _c suffix), import the module normally in Python.
#
#         Note that the Python routines operate on memoryview slices (compatible with np.arrays),
#         so they have slightly different parameters and return values when compared to the C routines.
#
#         Generally, the Python versions will allocate arrays for you, while the C versions expect you
#         to provide pointers to already malloc()'d memory (and explicit sizes).
#
# See the function docstrings and comments in the .pyx source for details.
#
# JJ 2016-11-07

##############################################################################################################
# Helpers
##############################################################################################################

cdef void distribute_items_c( int nitems, int ntasks, int* blocksizes, int* baseidxs ) nogil  # distribute work items across tasks, assuming equal load per item.

cdef void copygeneral_c( double* O, double* I, int nrows, int ncols ) nogil  # copy general square array
cdef void copysymmu_c( double* O, double* I, int nrows, int ncols ) nogil  # copy symmetric square array, upper triangle only

cdef void symmetrize_c( double* A, int nrows, int ncols ) nogil
cdef void msymmetrize_c( double* A, int nrows, int ncols, int nlhs ) nogil
cdef void msymmetrizep_c( double* A, int nrows, int ncols, int nlhs, int ntasks ) nogil

##############################################################################################################
# Preconditioning (scaling)
##############################################################################################################

# Scaling reduces the condition number of A, helping DGESV (general()) to give more correct digits.
#
# The return value is the number of iterations taken; always 1 for non-iterative algorithms.

# helpers
cdef void init_scaling_c( int nrows, int ncols, double* row_scale, double* col_scale ) nogil  # init all scaling factors to 1.0
cdef void apply_scaling_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil  # freeze the scaling by applying it in-place

# simple, fast methods; these destroy the possible symmetry of A
cdef int rescale_columns_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil
cdef int rescale_rows_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil
cdef int rescale_twopass_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil  # scale columns, then rows
cdef int rescale_dgeequ_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil

# symmetry-preserving methods (iterative)
cdef int rescale_ruiz2001_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil
cdef int rescale_scalgm_c( double* A, int nrows, int ncols, double* row_scale, double* col_scale ) nogil

##############################################################################################################
# Tridiagonal matrices
##############################################################################################################

cpdef int tridiag( double[::1] a, double[::1] b, double[::1] c, double[::1] x ) nogil except -1

##############################################################################################################
# Symmetric matrices
##############################################################################################################

cdef int symmetric2x2_c( double* A, double* b ) nogil except -1

cdef int symmetric_c( double* A, double* b, int n ) nogil except -1
cdef int symmetricfactor_c( double* A, int* ipiv, int n ) nogil except -1
cdef int symmetricfactored_c( double* A, int* ipiv, double* b, int n ) nogil except -1

cdef int symmetrics_c( double* A, double* b, int n, int nrhs ) nogil except -1
cdef int symmetricsp_c( double* A, double* b, int n, int nrhs, int ntasks ) nogil except -1

cdef int msymmetric_c( double* A, double* b, int n, int nlhs ) nogil except -1
cdef int msymmetricp_c( double* A, double* b, int n, int nlhs, int ntasks ) nogil except -1

cdef int msymmetricfactor_c( double* A, int* ipiv, int n, int nlhs ) nogil except -1
cdef int msymmetricfactored_c( double* A, int* ipiv, double* b, int n, int nlhs ) nogil except -1
cdef int msymmetricfactorp_c( double* A, int* ipiv, int n, int nlhs, int ntasks ) nogil except -1
cdef int msymmetricfactoredp_c( double* A, int* ipiv, double* b, int n, int nlhs, int ntasks ) nogil except -1

##############################################################################################################
# General matrices
##############################################################################################################

cdef int general2x2_c( double* A, double* b ) nogil except -1

cdef int general_c( double* A, double* b, int n ) nogil except -1
cdef int generalfactor_c( double* A, int* ipiv, int n ) nogil except -1
cdef int generalfactored_c( double* A, int* ipiv, double* b, int n ) nogil except -1

cdef int generals_c( double* A, double* b, int n, int nrhs ) nogil except -1
cdef int generalsp_c( double* A, double* b, int n, int nrhs, int ntasks ) nogil except -1

cdef int mgeneral_c( double* A, double* b, int n, int nlhs ) nogil except -1
cdef int mgeneralp_c( double* A, double* b, int n, int nlhs, int ntasks ) nogil except -1

cdef int mgeneralfactor_c( double* A, int* ipiv, int n, int nlhs ) nogil except -1
cdef int mgeneralfactored_c( double* A, int* ipiv, double* b, int n, int nlhs ) nogil except -1
cdef int mgeneralfactorp_c( double* A, int* ipiv, int n, int nlhs, int ntasks ) nogil except -1
cdef int mgeneralfactoredp_c( double* A, int* ipiv, double* b, int n, int nlhs, int ntasks ) nogil except -1

##############################################################################################################
# Other stuff
##############################################################################################################

cdef int svd_c( double* A, int m, int n, double* S ) nogil except -1




# -*- coding: utf-8 -*-
#
"""Setuptools-based setup script for WLSQM."""

from __future__ import division, print_function, absolute_import

#########################################################
# Config
#########################################################

# choose build type here
#
build_type="optimized"
#build_type="debug"


#########################################################
# Init
#########################################################

# check for Python 2.7 or later
# http://stackoverflow.com/questions/19534896/enforcing-python-version-in-setup-py
import sys
if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')

import os

from setuptools import setup
from setuptools.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    sys.exit("Cython not found. Cython is needed to build the extension modules for WLSQM.")


#########################################################
# Definitions
#########################################################

extra_compile_args_math_optimized    = ['-fopenmp', '-march=native', '-O2', '-msse', '-msse2', '-mfma', '-mfpmath=sse']
extra_compile_args_math_debug        = ['-fopenmp', '-march=native', '-O0', '-g']

extra_compile_args_nonmath_optimized = ['-O2']
extra_compile_args_nonmath_debug     = ['-O0', '-g']

extra_link_args_optimized    = ['-fopenmp']
extra_link_args_debug        = ['-fopenmp']


if build_type == 'optimized':
    my_extra_compile_args_math    = extra_compile_args_math_optimized
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_optimized
    my_extra_link_args            = extra_link_args_optimized
    debug = False
    print( "build configuration selected: optimized" )
else: # build_type == 'debug':
    my_extra_compile_args_math    = extra_compile_args_math_debug
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_debug
    my_extra_link_args            = extra_link_args_debug
    debug = True
    print( "build configuration selected: debug" )


#########################################################
# Long description
#########################################################

DESC="""WLSQM (Weighted Least SQuares Meshless) is a fast and accurate meshless least-squares interpolator for Python, implemented in Cython.

Given scalar data values on a set of points in 1D, 2D or 3D, WLSQM constructs a piecewise polynomial global surrogate model (a.k.a. response surface), using up to 4th order polynomials.

Use cases include response surface modeling, and computing space derivatives of data known only as values at discrete points in space. No grid or mesh is needed.

Any derivative of the model function (e.g. d2f/dxdy) can be easily evaluated, up to the order of the polynomial.

Sensitivity data of solution DOFs (on the data values at points other than the reference in the local neighborhood) can be optionally computed.

Performance-critical parts are implemented in Cython. LAPACK is used via SciPy's Cython-level bindings. OpenMP is used for parallelization over the independent local problems (also in the linear solver step).

This implementation is targeted for high performance in a single-node environment, such as a laptop. The main target is x86_64.
"""

#########################################################
# Helpers
#########################################################

my_include_dirs = ["."]  # IMPORTANT, see https://github.com/cython/cython/wiki/PackageHierarchy

def ext(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_nonmath
                    )
def ext_math(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_math,
                      extra_link_args=my_extra_link_args,
                      libraries=["m"]  # "m" links libm, the math library on unix-likes; see http://docs.cython.org/src/tutorial/external.html
                    )

# http://stackoverflow.com/questions/13628979/setuptools-how-to-make-package-contain-extra-data-folder-and-all-folders-inside
datadirs  = ("examples",)
dataexts  = (".py", ".pyx", ".pxd", ".c", ".sh", ".lyx", ".pdf")
datafiles = []
getext = lambda filename: os.path.splitext(filename)[1]
for datadir in datadirs:
    datafiles.extend( [(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
                       for root, dirs, files in os.walk(datadir)] )

datafiles.append( ('.', ["README.md", "LICENSE.md", "TODO.md", "CHANGELOG.md"]) )


#########################################################
# Utility modules
#########################################################

ext_module_ptrwrap       = ext(     "wlsqm.utils.ptrwrap")        # Pointer wrapper for Cython/Python integration
ext_module_lapackdrivers = ext_math("wlsqm.utils.lapackdrivers")  # Simple Python interface to LAPACK for solving many independent linear equation systems efficiently in parallel. Built on top of scipy.linalg.cython_lapack.

#########################################################
# WLSQM (Weighted Least SQuares Meshless method)
#########################################################

ext_module_defs     = ext(     "wlsqm.fitter.defs")      # definitions (named constants)
ext_module_infra    = ext(     "wlsqm.fitter.infra")     # memory allocation infrastructure
ext_module_impl     = ext_math("wlsqm.fitter.impl")      # low-level routines (implementation)
ext_module_polyeval = ext_math("wlsqm.fitter.polyeval")  # evaluation of Taylor expansions and general polynomials
ext_module_interp   = ext_math("wlsqm.fitter.interp")    # interpolation of fitted model
ext_module_simple   = ext_math("wlsqm.fitter.simple")    # simple API
ext_module_expert   = ext_math("wlsqm.fitter.expert")    # advanced API

#########################################################

# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
#
import ast
with file('wlsqm/__init__.py') as f:
    for line in f:
        if line.startswith('__version__'):
            version = ast.parse(line).body[0].value.s
            break
    else:
        version = '0.0.unknown'
        print( "WARNING: Version information not found, using placeholder '%s'" % (version) )


setup(
    name = "wlsqm",
    version = version,
    author = "Juha Jeronen",
    author_email = "juha.jeronen@jyu.fi",
    url = "https://github.com/Technologicat/python-wlsqm",

    description = "WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar data on 1D, 2D and 3D point clouds.",
    long_description = DESC,

    license = "BSD",
    platforms = ["Linux"],  # free-form text field; http://stackoverflow.com/questions/34994130/what-platforms-argument-to-setup-in-setup-py-does

    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: Console",
                    "Intended Audience :: Developers",
                    "Intended Audience :: Science/Research",
                    "License :: OSI Approved :: BSD License",
                    "Operating System :: POSIX :: Linux",
                    "Programming Language :: Cython",
                    "Programming Language :: Python",
                    "Programming Language :: Python :: 2",
                    "Programming Language :: Python :: 2.7",
                    "Programming Language :: Python :: 2 :: Only",
                    "Topic :: Scientific/Engineering",
                    "Topic :: Scientific/Engineering :: Mathematics",
                    "Topic :: Software Development :: Libraries",
                    "Topic :: Software Development :: Libraries :: Python Modules"
                  ],

    # 0.16 seems to be the first SciPy version that has cython_lapack.pxd. ( https://github.com/scipy/scipy/commit/ba438eab99ce8f55220a6ff652500f07dd6a547a )
    setup_requires = ["cython", "scipy (>=0.16)"],
    install_requires = ["numpy", "scipy (>=0.16)"],
    provides = ["wlsqm"],

    # same keywords as used as topics on GitHub
    keywords = ["numerical interpolation differentiation curve-fitting least-squares meshless numpy cython"],

    ext_modules = cythonize( [ ext_module_lapackdrivers,
                               ext_module_ptrwrap,
                               ext_module_defs,
                               ext_module_infra,
                               ext_module_impl,
                               ext_module_polyeval,
                               ext_module_interp,
                               ext_module_simple,
                               ext_module_expert ],

                             include_path = my_include_dirs,

                             gdb_debug = debug ),

    # Declare packages so that  python -m setup build  will copy .py files (especially __init__.py).
    packages = ["wlsqm", "wlsqm.utils", "wlsqm.fitter"],

    # Install also Cython headers so that other Cython modules can cimport ours
    # FIXME: force sdist, but sdist only, to keep the .pyx files (this puts them also in the bdist)
    package_data={'wlsqm.utils': ['*.pxd', '*.pyx'],  # note: paths relative to each package
                  'wlsqm.fitter': ['*.pxd', '*.pyx']},

    # Disable zip_safe, because:
    #   - Cython won't find .pxd files inside installed .egg, hard to compile libs depending on this one
    #   - dynamic loader may need to have the library unzipped to a temporary folder anyway (at import time)
    zip_safe = False,

    # Usage examples; not in a package
    data_files = datafiles
)



# -*- coding: utf-8 -*-
#
"""Performance benchmarking and usage examples for the wlsqm.lapackdrivers module.

JJ 2016-11-02
"""

from __future__ import division
from __future__ import absolute_import

import time
import math

import numpy as np
from numpy.linalg import solve as numpy_solve  # for comparison purposes

import pylab as pl

try:
    import wlsqm.utils.lapackdrivers as drivers
except ImportError:
    print "WLSQM not found; is it installed?"
    from sys import exit
    exit(1)

# from find_neighbors2.py
class SimpleTimer:
    def __init__(self, label="", n=None):
        self.label = label
        self.n     = n      # number of repetitions done inside the "with..." section (for averaging in timing info)

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, errtype, errvalue, traceback):
        dt         = time.time() - self.t0
        identifier = ("%s" % self.label) if len(self.label) else "time taken: "
        avg        = (", avg. %gs per run" % (dt/self.n)) if self.n is not None else ""
        print "%s%gs%s" % (identifier, dt, avg)

# from util.py
def f5(seq, idfun=None):
   """Uniqify a list (remove duplicates).

   This is the fast order-preserving uniqifier "f5" from
   http://www.peterbe.com/plog/uniqifiers-benchmark

   The list does not need to be sorted.

   The return value is the uniqified list.

   """
   # order preserving
   if idfun is None:
       def idfun(x): return x
   seen = {}
   result = []
   for item in seq:
       marker = idfun(item)
       # in old Python versions:
       # if seen.has_key(marker)
       # but in new ones:
       if marker in seen: continue
       seen[marker] = 1
       result.append(item)
   return result



def main():
#    # exact solution is (3/10, 2/5, 0)
#    A = np.array( ( (2., 1.,  3.),
#                    (2., 6.,  8.),
#                    (6., 8., 18.) ), dtype=np.float64, order='F' )
#    b = np.array(   (1., 3., 5.),    dtype=np.float64 )

#    # symmetric matrix for testing symmetric solver
#    A = np.array( ( (2., 1.,  3.),
#                    (1., 6.,  8.),
#                    (3., 8., 18.) ), dtype=np.float64, order='F' )
#    b = np.array(   (1., 3., 5.),    dtype=np.float64 )

    # random matrix
    n = 5
    A = np.random.sample( (n,n) )
    A = np.array( A, dtype=np.float64, order='F' )
    drivers.symmetrize( A )  # fast Cython implementation of  A = 0.5 * (A + A.T)
    b = np.random.sample( (n,) )

    # test that it works

    x = numpy_solve(A, b)
    print "NumPy:", x

    A2 = A.copy(order='F')
    x2 = b.copy()
    drivers.symmetric(A2, x2)
    print "dsysv:", x2

    A3 = A.copy(order='F')
    x3 = b.copy()
    drivers.general(A3, x3)
    print "dgesv:", x3

    assert (np.abs(x - x3) < 1e-10).all(), "Something went wrong, solutions do not match"  # check general solver first
    assert (np.abs(x - x2) < 1e-10).all(), "Something went wrong, solutions do not match"  # then check symmetric solver


    # test performance

    # for verification only - very slow (serial only!)
    use_numpy = True

    # parallel processing
    ntasks = 8

#    # overview, somewhat fast but not very accurate
#    sizes = f5( map( lambda x: int(x), np.ceil(3*np.logspace(0, 3, 21, dtype=int)) ) )
#    reps = map( lambda x: int(x), 10.**(4 - np.log10(sizes)) )

#    # "large n"
#    sizes = f5( map( lambda x: int(x), np.ceil(3*np.logspace(2, 3, 21, dtype=int)) ) )
#    reps = map( lambda x: int(x), 10.**(5 - np.log10(sizes)) )

    # "small n" (needs more repetitions to eliminate noise from other running processes since a single solve is very fast)
    sizes = f5( map( lambda x: int(x), np.ceil(3*np.logspace(0, 2, 21, dtype=int)) ) )
    reps = map( lambda x: int(x), 10.**(6 - np.log10(sizes)) )

    print "performance test: %d tasks, sizes %s" % (ntasks, sizes)

    results1 = np.empty( (len(sizes),), dtype=np.float64 )
    results2 = np.empty( (len(sizes),), dtype=np.float64 )
    results3 = np.empty( (len(sizes),), dtype=np.float64 )
    results4 = np.empty( (len(sizes),), dtype=np.float64 )
    results5 = np.empty( (len(sizes),), dtype=np.float64 )
    results6 = np.empty( (len(sizes),), dtype=np.float64 )
    results7 = np.empty( (len(sizes),), dtype=np.float64 )

#    # many LHS (completely independent problems)
#    n = 5
#    reps=int(1e5)
#    A = np.random.sample( (n,n,reps) )
#    A = 0.5 * (A + A.transpose(1,0,2))  # symmetrize
#    A = np.array( A, dtype=np.float64, order='F' )
#    b = np.random.sample( (n,reps) )
#    b = np.array( b, dtype=np.float64, order='F' )
#    with SimpleTimer(label="msymmetric ", n=reps) as s:
#        drivers.msymmetricp(A, b, ntasks)
#    with SimpleTimer(label="mgeneral ", n=reps) as s:
#        drivers.mgeneralp(A, b, ntasks)


    for j,item in enumerate(zip(sizes,reps)):
        n,r = item
        print "testing size %d, reps = %d" % (n, r)

        # same LHS, many different RHS

        print "    prep same LHS, many RHS..."

        A = np.random.sample( (n,n) )
        # symmetrize
#        A *= 0.5
#        A += A.T  # not sure if this works
        A = np.array( A, dtype=np.float64, order='F' )
#        A = 0.5 * (A + A.T)  # symmetrize
        drivers.symmetrize(A)
        b = np.random.sample( (n,r) )
        b = np.array( b, dtype=np.float64, order='F' )

        print "    solve:"

#        # for verification only - very slow (Python loop, serial!)
#        if use_numpy:
#            t0 = time.time()
#            x = np.empty( (n,r), dtype=np.float64 )
#            for k in xrange(r):
#                x[:,k] = numpy_solve(A, b[:,k])
#            results1[j] = (time.time() - t0) / r

        print "        symmetricsp"
        t0 = time.time()
        A2 = A.copy(order='F')
        x2 = b.copy(order='F')
        drivers.symmetricsp(A2, x2, ntasks)
        results2[j] = (time.time() - t0) / r

        print "        generalsp"
        t0 = time.time()
        A3 = A.copy(order='F')
        x3 = b.copy(order='F')
        drivers.generalsp(A3, x3, ntasks)
        results3[j] = (time.time() - t0) / r

        # different LHS for each problem

        print "    prep independent problems..."

        A = np.random.sample( (n,n,r) )
        # symmetrize
#        A *= 0.5
#        A += A.transpose(1,0,2)  # this doesn't work
        A = np.array( A, dtype=np.float64, order='F' )
#        A = 0.5 * (A + A.transpose(1,0,2))
        drivers.msymmetrizep(A, ntasks)
        b = np.random.sample( (n,r) )
        b = np.array( b, dtype=np.float64, order='F' )

        print "    solve:"

        # for verification only - very slow (Python loop, serial!)
        if use_numpy:
            print "        NumPy"
            t0 = time.time()
            x = np.empty( (n,r), dtype=np.float64, order='F' )
            for k in xrange(r):
                x[:,k] = numpy_solve(A[:,:,k], b[:,k])
            results1[j] = (time.time() - t0) / r

        print "        msymmetricp"
        t0 = time.time()
        A2 = A.copy(order='F')
        x2 = b.copy(order='F')
        drivers.msymmetricp(A2, x2, ntasks)
        results4[j] = (time.time() - t0) / r

        print "        mgeneralp"
        t0 = time.time()
        A3 = A.copy(order='F')
        x3 = b.copy(order='F')
        drivers.mgeneralp(A3, x3, ntasks)
        results5[j] = (time.time() - t0) / r

        print "        msymmetricfactorp & msymmetricfactoredp"  # factor once, then it is possible to solve multiple times (although we now test only once)
        t0 = time.time()
        ipiv = np.empty( (n,r), dtype=np.int32, order='F' )
        fact = A.copy(order='F')
        x4   = b.copy(order='F')
        drivers.msymmetricfactorp( fact, ipiv, ntasks )
        drivers.msymmetricfactoredp( fact, ipiv, x4, ntasks )
        results6[j] = (time.time() - t0) / r

        print "        mgeneralfactorp & mgeneralfactoredp"  # factor once, then it is possible to solve multiple times (although we now test only once)
        t0 = time.time()
        ipiv = np.empty( (n,r), dtype=np.int32, order='F' )
        fact = A.copy(order='F')
        x5   = b.copy(order='F')
        drivers.mgeneralfactorp( fact, ipiv, ntasks )
        drivers.mgeneralfactoredp( fact, ipiv, x5, ntasks )
        results7[j] = (time.time() - t0) / r

        if use_numpy:
#            print np.max(np.abs(x - x3))  # DEBUG
#            print np.max(np.abs(x - x5))  # DEBUG
            print np.max(np.abs(x2 - x4))  # DEBUG
            assert (np.abs(x - x5) < 1e-10).all(), "Something went wrong, solutions do not match"  # check general solver first
            assert (np.abs(x - x3) < 1e-10).all(), "Something went wrong, solutions do not match"  # check general solver
#            assert (np.abs(x - x2) < 1e-5).all(), "Something went wrong, solutions do not match"  # doesn't make sense to compare, DSYSV is more accurate for badly conditioned symmetric matrices
            assert (np.abs(x2 - x4) < 1e-7).all(), "Something went wrong, solutions do not match"  # check symmetric solvers against each other
                                                                                                   # (not exactly the same algorithm (DSYTRS2 vs. DSYTRS), so there may be slight deviation)


# old, serial only
#
#    for j,item in enumerate(zip(sizes,reps)):
#        n,r = item
#        print "testing size %d, reps = %d" % (n, r)
#
#        A = np.random.sample( (n,n) )
#        A = 0.5 * (A + A.T)  # symmetrize
#        A = np.array( A, dtype=np.float64, order='F' )
#        b = np.random.sample( (n,) )
#
#        t0 = time.time()
#        for k in xrange(r):
#            x = numpy_solve(A, b)
#        results1[j] = (time.time() - t0) / r
#
#        t0 = time.time()
#        for k in xrange(r):
#            A2 = A.copy(order='F')
#            x2 = b.copy()
#            drivers.symmetric(A2, x2)
#        results2[j] = (time.time() - t0) / r
#
#        t0 = time.time()
#        for k in xrange(r):
#            A3 = A.copy(order='F')
#            x3 = b.copy()
#            drivers.general(A3, x3)
#        results3[j] = (time.time() - t0) / r


    # visualize

    pl.figure(1)
    pl.clf()
    if use_numpy:
        pl.loglog(sizes, results1, 'k-', label='NumPy')
    pl.loglog(sizes, results2, 'b--', label='dsysv, same LHS, many RHS')
    pl.loglog(sizes, results3, 'b-',  label='dgesv, same LHS, many RHS')
    pl.loglog(sizes, results4, 'r--', label='dsysv, independent problems')
    pl.loglog(sizes, results5, 'r-',  label='dgesv, independent problems')
    pl.loglog(sizes, results6, 'g--', label='dsytrf+dsytrs, independent problems')
    pl.loglog(sizes, results7, 'g-',  label='dgetrf+dgetrs, independent problems')
    pl.xlabel('n')
    pl.ylabel('t')
    pl.title('Average time per problem instance, %d parallel tasks' % (ntasks))
    pl.axis('tight')
    pl.grid(b=True, which='both')
    pl.legend(loc='best')

    pl.savefig('figure1_latest.pdf')


if __name__ == '__main__':
    main()
    pl.show()




# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# Evaluation of Taylor expansions and general polynomials up to 4th order in 1D, 2D and 3D.
#
# C API definitions.
#
# JJ 2016-12-09

from cython cimport view

cdef int taylor_3D( int order, double* fi, double xi, double yi, double zi, double[::view.generic,::view.contiguous] x, double* out ) nogil
cdef int general_3D( int order, double* fi, double xi, double yi, double zi, double[::view.generic,::view.contiguous] x, double* out ) nogil

cdef int taylor_2D( int order, double* fi, double xi, double yi, double[::view.generic,::view.contiguous] x, double* out ) nogil
cdef int general_2D( int order, double* fi, double xi, double yi, double[::view.generic,::view.contiguous] x, double* out ) nogil

cdef int taylor_1D( int order, double* fi, double xi, double[::view.generic] x, double* out ) nogil
cdef int general_1D( int order, double* fi, double xi, double[::view.generic] x, double* out ) nogil




# -*- coding: utf-8 -*-
"""A minimal usage example for ExpertSolver.

JJ 2017-03-28
"""

from __future__ import division

import numpy as np

import scipy.spatial.ckdtree

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d

import wlsqm


def project_onto_regular_grid_2D(x, F, nvis=101, fit_order=1, nk=10):
    """Project scalar data from a 2D point cloud onto a regular grid.

Useful for plotting. Uses the WLSQM meshless method.

The bounding box of the x data is automatically used as the bounds of the generated regular grid.

Parameters:
    x : rank-2 array, dtype np.float64
        Point cloud, one point per row. x[i,:] = (xi,yi)

    F : rank-1 array, dtype np.float64
        The corresponding function values. F[i] = F( x[i,:] )

    nvis : int
        Number of points per axis in the generated regular grid.

    fit_order : int
        Order of the surrogate polynomial, one of [0,1,2,3,4].

    nk : int
        Number of nearest neighbors to use for fitting the model.

Return value:
    tuple (X,Y,Z)
        where
        X,Y are rank-2 meshgrid arrays representing the generated regular grid, and
        Z is an array of the same shape, containing the corresponding data values.

"""
    # Form the neighborhoods.

    # index the input points for fast searching
    tree = scipy.spatial.cKDTree( data=x )

    # Find max_nk nearest neighbors of each input point.
    #
    # The +1 is for the point itself, since it is always the nearest to itself.
    #
    # (cKDTree.query() supports querying for arbitrary x; here we just set these x as the same as the points in the tree.)
    #
    dd,ii = tree.query( x, 1 + nk )

    # Take only the neighbors of points[i], excluding the point itself.
    #
    ii = ii[:,1:]  # points[ ii[i,k] ] is the kth nearest neighbor of points[i]. Shape of ii is (npoints, nk).

    # neighbor point indices (pointing to rows in x[]); typecast to int32
    hoods = np.array( ii, dtype=np.int32 )

    npoints  = x.shape[0]
    nk_array = nk * np.ones( (npoints,), dtype=np.int32 )  # number of neighbors, i.e. nk_array[i] is the number of actually used columns in hoods[i,:]

    # Construct the model by least-squares fitting
    #
    fit_order_array = fit_order            * np.ones( (npoints,), dtype=np.int32 )
    knowns_array    = wlsqm.b2_F           * np.ones( (npoints,), dtype=np.int64 )  # bitmask! wlsqm.b*
    wm_array        = wlsqm.WEIGHT_UNIFORM * np.ones( (npoints,), dtype=np.int32 )
    solver = wlsqm.ExpertSolver( dimension=2,
                                 nk=nk_array,
                                 order=fit_order_array,
                                 knowns=knowns_array,
                                 weighting_method=wm_array,
                                 algorithm=wlsqm.ALGO_BASIC,
                                 do_sens=False,
                                 max_iter=10,  # must be an int even though this parameter is not used in ALGO_BASIC mode
                                 ntasks=8,
                                 debug=False )

    no = wlsqm.number_of_dofs( dimension=2, order=fit_order )
    fi = np.empty( (npoints,no), dtype=np.float64 )
    fi[:,0] = F  # fi[i,0] contains the function value at point x[i,:]

    solver.prepare( xi=x, xk=x[hoods] )  # generate problem matrices from the geometry of the point cloud
    solver.solve( fk=fi[hoods,0], fi=fi, sens=None )  # compute least-squares fit to data


    # generate the regular grid for output
    #
    xx  = np.linspace( np.min(x[:,0]), np.max(x[:,0]), nvis )
    yy  = np.linspace( np.min(x[:,1]), np.max(x[:,1]), nvis )
    X,Y = np.meshgrid(xx,yy)

    # make a flat list of grid points (rank-2 array, one point per row)
    #
    Xlin = np.reshape(X, -1)
    Ylin = np.reshape(Y, -1)
    xout = np.empty( (len(Xlin), 2), dtype=np.float64 )
    xout[:,0] = Xlin
    xout[:,1] = Ylin

    # Using the model, interpolate onto the regular grid
    #
    solver.prep_interpolate()  # prepare global model
    Z,mi = solver.interpolate( xout, mode='nearest' )  # use the nearest local model; fast, surprisingly accurate
                                                       # if a reasonable number of points (and continuous-looking
                                                       # although technically has jumps over Voronoi cell boundaries)
    # when mode="nearest", "mi" is an array containing the index of the local model (which belongs to x[mi,:]) used for each evaluation

    return (X, Y, np.reshape( Z, X.shape ))


def plot_wireframe( data, figno=None ):
    """Make and label a wireframe plot.

Parameters:
    data : dict
        key   : "x","y","z"
        value : tuple (rank-2 array in meshgrid format, axis label)

Return value:
    ax
        The Axes3D object that was used for plotting.
"""
    # http://matplotlib.org/mpl_toolkits/mplot3d/tutorial.html
    fig = plt.figure(figno)

    # Axes3D has a tendency to underestimate how much space it needs; it draws its labels
    # outside the window area in certain orientations.
    #
    # This causes the labels to be clipped, which looks bad. We prevent this by creating the axes
    # in a slightly smaller rect (leaving a margin). This way the labels will show - outside the Axes3D,
    # but still inside the figure window.
    #
    # The final touch is to set the window background to a matching white, so that the
    # background of the figure appears uniform.
    #
    fig.patch.set_color( (1,1,1) )
    fig.patch.set_alpha( 1.0 )
    x0y0wh = [ 0.02, 0.02, 0.96, 0.96 ]  # left, bottom, width, height      (here as fraction of subplot area)

    ax = mpl_toolkits.mplot3d.axes3d.Axes3D(fig, rect=x0y0wh)

    X,xlabel = data["x"]
    Y,ylabel = data["y"]
    Z,zlabel = data["z"]
    ax.plot_wireframe( X, Y, Z )

    ax.view_init(34, -40)
    ax.axis('tight')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    ax.set_title(zlabel)

    return ax


def main():
    x = np.random.random( (1000, 2) )                # point cloud (no mesh topology!)
    F = np.sin(np.pi*x[:,0]) * np.cos(np.pi*x[:,1])  # function values on the point cloud
    X,Y,Z = project_onto_regular_grid_2D(x, F, fit_order=2, nk=30)
    plot_wireframe( {"x" : (X, r"$x$"),
                     "y" : (Y, r"$y$"),
                     "z" : (Z, r"$f(x,y)$")} )

if __name__ == '__main__':
    main()
    plt.show()
